@@ -15,7 +15,6 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, see <https://www.gnu.org/licenses/>.
- * SPDX-License-Identifier: GPL-3.0-or-later
  */
 
 #include <config.h>
@@ -39,7 +38,6 @@
 #include <fcntl.h> /* for setmode() */
 #endif
 
-#define INCLUDED_BY_MAIN_MODULE 1
 #include "agent.h"
 #include "../common/i18n.h"
 #include "../common/get-passphrase.h"
@@ -146,11 +144,9 @@ my_strusage (int level)
   const char *p;
   switch (level)
     {
-    case  9: p = "GPL-3.0-or-later"; break;
     case 11: p = "gpg-protect-tool (" GNUPG_NAME ")";
       break;
     case 13: p = VERSION; break;
-    case 14: p = GNUPG_DEF_COPYRIGHT_LINE; break;
     case 17: p = PRINTABLE_OS_NAME; break;
     case 19: p = _("Please report bugs to <@EMAIL@>.\n"); break;
 
@@ -278,7 +274,7 @@ read_file (const char *fname, size_t *r_length)
     {
       struct stat st;
 
-      fp = gnupg_fopen (fname, "rb");
+      fp = fopen (fname, "rb");
       if (!fp)
         {
           log_error ("can't open '%s': %s\n", fname, strerror (errno));
@@ -319,11 +315,6 @@ read_key (const char *fname)
   buf = read_file (fname, &buflen);
   if (!buf)
     return NULL;
-  if (buflen >= 4 && !memcmp (buf, "Key:", 4))
-    {
-      log_error ("Extended key format is not supported by this tool\n");
-      return NULL;
-    }
   key = make_canonical (fname, buf, buflen);
   xfree (buf);
   return key;
@@ -554,6 +545,7 @@ show_keygrip (const char *fname)
   putchar ('\n');
 }
 
+
 
 
 
@@ -579,8 +571,8 @@ main (int argc, char **argv )
 
   pargs.argc = &argc;
   pargs.argv = &argv;
-  pargs.flags= ARGPARSE_FLAG_KEEP;
-  while (gnupg_argparse (NULL, &pargs, opts))
+  pargs.flags=  1;  /* (do not remove the args) */
+  while (arg_parse (&pargs, opts) )
     {
       switch (pargs.r_opt)
         {
@@ -610,8 +602,6 @@ main (int argc, char **argv )
         default: pargs.err = ARGPARSE_PRINT_ERROR; break;
 	}
     }
-  gnupg_argparse (NULL, &pargs, NULL);  /* Release internal state.  */
-
   if (log_get_errorcount (0))
     exit (2);
 
@@ -759,9 +749,8 @@ agent_key_available (const unsigned char *grip)
 }
 
 char *
-agent_get_cache (ctrl_t ctrl, const char *key, cache_mode_t cache_mode)
+agent_get_cache (const char *key, cache_mode_t cache_mode)
 {
-  (void)ctrl;
   (void)key;
   (void)cache_mode;
   return NULL;
@@ -809,19 +798,12 @@ agent_askpin (ctrl_t ctrl,
  * to stdout. */
 int
 agent_write_private_key (const unsigned char *grip,
-                         const void *buffer, size_t length, int force,
-                         time_t timestamp,
-                         const char *serialno, const char *keyref,
-                         const char *dispserialno)
+                         const void *buffer, size_t length, int force)
 {
   char hexgrip[40+4+1];
   char *p;
 
   (void)force;
-  (void)timestamp;
-  (void)serialno;
-  (void)keyref;
-  (void)dispserialno;
 
   bin2hex (grip, 20, hexgrip);
   strcpy (hexgrip+40, ".key");
