@@ -1,7 +1,7 @@
 /* encrypt.c - Main encryption driver
  * Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005,
  *               2006, 2009 Free Software Foundation, Inc.
- * Copyright (C) 2016, 2023 g10 Code GmbH
+ * Copyright (C) 2016 g10 Code GmbH
  *
  * This file is part of GnuPG.
  *
@@ -17,7 +17,6 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, see <https://www.gnu.org/licenses/>.
- * SPDX-License-Identifier: GPL-3.0-or-later
  */
 
 #include <config.h>
@@ -48,12 +47,12 @@ static int write_pubkey_enc_from_list (ctrl_t ctrl,
 
 /****************
  * Encrypt FILENAME with only the symmetric cipher.  Take input from
- * stdin if FILENAME is NULL.  If --force-aead is used we use an SKESK.
+ * stdin if FILENAME is NULL.
  */
 int
 encrypt_symmetric (const char *filename)
 {
-  return encrypt_simple( filename, 1, opt.force_aead);
+  return encrypt_simple( filename, 1, 0 );
 }
 
 
@@ -68,324 +67,99 @@ encrypt_store (const char *filename)
 }
 
 
-/* Create and setup a DEK structure and print approriate warnings.
- * PK_LIST gives the list of public keys.  Always returns a DEK.  The
- * actual session needs to be added later.  */
-static DEK *
-create_dek_with_warnings (pk_list_t pk_list)
+/* *SESKEY contains the unencrypted session key ((*SESKEY)->KEY) and
+   the algorithm that will be used to encrypt the contents of the SED
+   packet ((*SESKEY)->ALGO).  If *SESKEY is NULL, then a random
+   session key that is appropriate for DEK->ALGO is generated and
+   stored there.
+
+   Encrypt that session key using DEK and store the result in ENCKEY,
+   which must be large enough to hold (*SESKEY)->KEYLEN + 1 bytes.  */
+void
+encrypt_seskey (DEK *dek, DEK **seskey, byte *enckey)
 {
-  DEK *dek;
+  gcry_cipher_hd_t hd;
+  byte buf[33];
 
-  dek = xmalloc_secure_clear (sizeof *dek);
-  if (!opt.def_cipher_algo)
+  log_assert ( dek->keylen <= 32 );
+  if (!*seskey)
     {
-      /* Try to get it from the prefs.  */
-      dek->algo = select_algo_from_prefs (pk_list, PREFTYPE_SYM, -1, NULL);
-      if (dek->algo == -1)
-        {
-          /* If does not make sense to fallback to the rfc4880
-           * required 3DES if we will reject that algo later.  Thus we
-           * fallback to AES anticipating RFC4880bis rules.  */
-          if (opt.flags.allow_old_cipher_algos)
-            dek->algo = CIPHER_ALGO_3DES;
-          else
-            dek->algo = CIPHER_ALGO_AES;
-        }
-
-      /* In case 3DES has been selected, print a warning if any key
-       * does not have a preference for AES.  This should help to
-       * indentify why encrypting to several recipients falls back to
-       * 3DES. */
-      if (opt.verbose && dek->algo == CIPHER_ALGO_3DES)
-        warn_missing_aes_from_pklist (pk_list);
-    }
-  else
-    {
-      if (!opt.expert
-          && (select_algo_from_prefs (pk_list, PREFTYPE_SYM,
-                                      opt.def_cipher_algo, NULL)
-              != opt.def_cipher_algo))
-        {
-          log_info(_("WARNING: forcing symmetric cipher %s (%d)"
-                     " violates recipient preferences\n"),
-                   openpgp_cipher_algo_name (opt.def_cipher_algo),
-                   opt.def_cipher_algo);
-        }
-
-      dek->algo = opt.def_cipher_algo;
-    }
-
-  return dek;
-}
-
-
-/* Check whether all encryption keys are compliant with the current
- * mode and issue respective status lines.  DEK has the info about the
- * session key and PK_LIST the list of public keys.  */
-static gpg_error_t
-check_encryption_compliance (DEK *dek, pk_list_t pk_list)
-{
-  gpg_error_t err = 0;
-  pk_list_t pkr;
-  int compliant;
-
-  /* First check whether we should use the algo at all.  */
-  if (openpgp_cipher_blocklen (dek->algo) < 16
-      && !opt.flags.allow_old_cipher_algos)
-    {
-      log_error (_("cipher algorithm '%s' may not be used for encryption\n"),
-		 openpgp_cipher_algo_name (dek->algo));
-      if (!opt.quiet)
-        log_info (_("(use option \"%s\" to override)\n"),
-                  "--allow-old-cipher-algos");
-      err = gpg_error (GPG_ERR_CIPHER_ALGO);
-      goto leave;
-    }
-
-  /* Now check the compliance.  */
-  if (! gnupg_cipher_is_allowed (opt.compliance, 1, dek->algo,
-                                 GCRY_CIPHER_MODE_CFB))
-    {
-      log_error (_("cipher algorithm '%s' may not be used in %s mode\n"),
-		 openpgp_cipher_algo_name (dek->algo),
-		 gnupg_compliance_option_string (opt.compliance));
-      err = gpg_error (GPG_ERR_CIPHER_ALGO);
-      goto leave;
-    }
-
-  if (!gnupg_rng_is_compliant (opt.compliance))
-    {
-      err = gpg_error (GPG_ERR_FORBIDDEN);
-      log_error (_("%s is not compliant with %s mode\n"),
-                 "RNG",
-                 gnupg_compliance_option_string (opt.compliance));
-      write_status_error ("random-compliance", err);
-      goto leave;
-    }
-
-  /* From here on we only test for CO_DE_VS - if we ever want to
-   * return other compliance mode values we need to change this to
-   * loop over all those values.  */
-  compliant = gnupg_gcrypt_is_compliant (CO_DE_VS);
-
-  if (!gnupg_cipher_is_compliant (CO_DE_VS, dek->algo, GCRY_CIPHER_MODE_CFB))
-    compliant = 0;
-
-  for (pkr = pk_list; pkr; pkr = pkr->next)
-    {
-      PKT_public_key *pk = pkr->pk;
-      unsigned int nbits = nbits_from_pk (pk);
-
-      if (!gnupg_pk_is_compliant (opt.compliance, pk->pubkey_algo, 0,
-                                  pk->pkey, nbits, NULL))
-        log_info (_("WARNING: key %s is not suitable for encryption"
-                    " in %s mode\n"),
-                  keystr_from_pk (pk),
-                  gnupg_compliance_option_string (opt.compliance));
-
-      if (compliant
-          && !gnupg_pk_is_compliant (CO_DE_VS, pk->pubkey_algo, 0, pk->pkey,
-                                     nbits, NULL))
-        compliant = 0; /* Not compliant - reset flag.  */
-    }
-
-  /* If we are compliant print the status for de-vs compliance.  */
-  if (compliant)
-    write_status_strings (STATUS_ENCRYPTION_COMPLIANCE_MODE,
-                          gnupg_status_compliance_flag (CO_DE_VS),
-                          NULL);
-
-  /* Check whether we should fail the operation.  */
-  if (opt.flags.require_compliance
-      && opt.compliance == CO_DE_VS
-      && !compliant)
-    {
-      compliance_failure ();
-      err = gpg_error (GPG_ERR_FORBIDDEN);
-      goto leave;
-    }
-
- leave:
-  return err;
-}
-
-
-/* Encrypt a session key using DEK and store a pointer to the result
- * at R_ENCKEY and its length at R_ENCKEYLEN.
- *
- * R_SESKEY points to the unencrypted session key (.KEY, .KEYLEN) and
- * the algorithm that will be used to encrypt the contents of the
- * SKESK packet (.ALGO).  If R_SESKEY points to NULL, then a random
- * session key that is appropriate for DEK->ALGO is generated and
- * stored at R_SESKEY.  If AEAD_ALGO is not 0 the given AEAD algorithm
- * is used for encryption.
- */
-static gpg_error_t
-encrypt_seskey (DEK *dek, aead_algo_t aead_algo,
-                DEK **r_seskey, void **r_enckey, size_t *r_enckeylen)
-{
-  gpg_error_t err;
-  gcry_cipher_hd_t hd = NULL;
-  byte *buf = NULL;
-  DEK *seskey;
-
-  *r_enckey = NULL;
-  *r_enckeylen = 0;
-
-  if (*r_seskey)
-    seskey = *r_seskey;
-  else
-    {
-      seskey = xtrycalloc (1, sizeof(DEK));
-      if (!seskey)
-        {
-          err = gpg_error_from_syserror ();
-          goto leave;
-        }
-      seskey->algo = dek->algo;
-      make_session_key (seskey);
+      *seskey=xmalloc_clear(sizeof(DEK));
+      (*seskey)->algo=dek->algo;
+      make_session_key(*seskey);
       /*log_hexdump( "thekey", c->key, c->keylen );*/
     }
 
+  /* The encrypted session key is prefixed with a one-octet algorithm id.  */
+  buf[0] = (*seskey)->algo;
+  memcpy( buf + 1, (*seskey)->key, (*seskey)->keylen );
 
-  if (aead_algo)
-    {
-      unsigned int noncelen;
-      enum gcry_cipher_modes ciphermode;
-      byte ad[4];
-
-      err = openpgp_aead_algo_info (aead_algo, &ciphermode, &noncelen);
-      if (err)
-        goto leave;
-
-      /* Allocate space for the nonce, the key, and the authentication
-       * tag (16).  */
-      buf = xtrymalloc_secure (noncelen + seskey->keylen + 16);
-      if (!buf)
-        {
-          err = gpg_error_from_syserror ();
-          goto leave;
-        }
-
-      gcry_randomize (buf, noncelen, GCRY_STRONG_RANDOM);
-
-      err = openpgp_cipher_open (&hd, dek->algo,
-                                 ciphermode, GCRY_CIPHER_SECURE);
-      if (!err)
-        err = gcry_cipher_setkey (hd, dek->key, dek->keylen);
-      if (!err)
-        err = gcry_cipher_setiv (hd, buf, noncelen);
-      if (err)
-        goto leave;
-
-      ad[0] = (0xc0 | PKT_SYMKEY_ENC);
-      ad[1] = 5;
-      ad[2] = dek->algo;
-      ad[3] = aead_algo;
-      err = gcry_cipher_authenticate (hd, ad, 4);
-      if (err)
-        goto leave;
-
-      memcpy (buf + noncelen, seskey->key, seskey->keylen);
-      gcry_cipher_final (hd);
-      err = gcry_cipher_encrypt (hd, buf + noncelen, seskey->keylen, NULL,0);
-      if (err)
-        goto leave;
-      err = gcry_cipher_gettag (hd, buf + noncelen + seskey->keylen, 16);
-      if (err)
-        goto leave;
-      *r_enckeylen = noncelen + seskey->keylen + 16;
-      *r_enckey = buf;
-      buf = NULL;
-    }
-  else
-    {
-      /* In the old version 4 SKESK the encrypted session key is
-       * prefixed with a one-octet algorithm id.  */
-      buf = xtrymalloc_secure (1 + seskey->keylen);
-      if (!buf)
-        {
-          err = gpg_error_from_syserror ();
-          goto leave;
-        }
-      buf[0] = seskey->algo;
-      memcpy (buf + 1, seskey->key, seskey->keylen);
-
-      err = openpgp_cipher_open (&hd, dek->algo, GCRY_CIPHER_MODE_CFB, 1);
-      if (!err)
-        err = gcry_cipher_setkey (hd, dek->key, dek->keylen);
-      if (!err)
-        err = gcry_cipher_setiv (hd, NULL, 0);
-      if (!err)
-        err = gcry_cipher_encrypt (hd, buf, seskey->keylen + 1, NULL, 0);
-      if (err)
-        goto leave;
-      *r_enckeylen = seskey->keylen + 1;
-      *r_enckey = buf;
-      buf = NULL;
-    }
-
-  /* Return the session key in case we allocated it.  */
-  *r_seskey = seskey;
-  seskey = NULL;
-
- leave:
+  /* We only pass already checked values to the following function,
+     thus we consider any failure as fatal.  */
+  if (openpgp_cipher_open (&hd, dek->algo, GCRY_CIPHER_MODE_CFB, 1))
+    BUG ();
+  if (gcry_cipher_setkey (hd, dek->key, dek->keylen))
+    BUG ();
+  gcry_cipher_setiv (hd, NULL, 0);
+  gcry_cipher_encrypt (hd, buf, (*seskey)->keylen + 1, NULL, 0);
   gcry_cipher_close (hd);
-  if (seskey != *r_seskey)
-    xfree (seskey);
-  xfree (buf);
-  return err;
+
+  memcpy( enckey, buf, (*seskey)->keylen + 1 );
+  wipememory( buf, sizeof buf ); /* burn key */
 }
 
 
-/* Return the AEAD algo if we shall use AEAD mode.  Returns 0 if AEAD
- * shall not be used.  */
-aead_algo_t
-use_aead (pk_list_t pk_list, int algo)
-{
-  int can_use;
-
-  can_use = openpgp_cipher_get_algo_blklen (algo) == 16;
-
-  /* With --force-aead we want AEAD.  */
-  if (opt.force_aead)
-    {
-      if (!can_use)
-        {
-          log_info ("Warning: request to use OCB ignored for cipher '%s'\n",
-                    openpgp_cipher_algo_name (algo));
-          return 0;
-        }
-      return AEAD_ALGO_OCB;
-    }
-
-  /* AEAD does only work with 128 bit cipher blocklength.  */
-  if (!can_use)
-    return 0;
-
-  /* Note the user which keys have no AEAD feature flag set.  */
-  if (opt.verbose)
-    warn_missing_aead_from_pklist (pk_list);
-
-  /* If all keys support AEAD we can use it.  */
-  return select_aead_from_pklist (pk_list);
-}
-
-
-/* Shall we use the MDC?  Yes - unless rfc-2440 compatibility is
- * requested. */
+/* We try very hard to use a MDC */
 int
 use_mdc (pk_list_t pk_list,int algo)
 {
-  (void)pk_list;
-  (void)algo;
-
-  /* RFC-2440 don't has MDC - this is the only way to create a legacy
-   * non-MDC encryption packet.  */
+  /* RFC-2440 don't has MDC */
   if (RFC2440)
     return 0;
 
-  return 1; /* In all other cases we use the MDC */
+  /* --force-mdc overrides --disable-mdc */
+  if(opt.force_mdc)
+    return 1;
+
+  if(opt.disable_mdc)
+    return 0;
+
+  /* Do the keys really support MDC? */
+
+  if(select_mdc_from_pklist(pk_list))
+    return 1;
+
+  /* The keys don't support MDC, so now we do a bit of a hack - if any
+     of the AESes or TWOFISH are in the prefs, we assume that the user
+     can handle a MDC.  This is valid for PGP 7, which can handle MDCs
+     though it will not generate them.  2440bis allows this, by the
+     way. */
+
+  if(select_algo_from_prefs(pk_list,PREFTYPE_SYM,
+			    CIPHER_ALGO_AES,NULL)==CIPHER_ALGO_AES)
+    return 1;
+
+  if(select_algo_from_prefs(pk_list,PREFTYPE_SYM,
+			    CIPHER_ALGO_AES192,NULL)==CIPHER_ALGO_AES192)
+    return 1;
+
+  if(select_algo_from_prefs(pk_list,PREFTYPE_SYM,
+			    CIPHER_ALGO_AES256,NULL)==CIPHER_ALGO_AES256)
+    return 1;
+
+  if(select_algo_from_prefs(pk_list,PREFTYPE_SYM,
+			    CIPHER_ALGO_TWOFISH,NULL)==CIPHER_ALGO_TWOFISH)
+    return 1;
+
+  /* Last try.  Use MDC for the modern ciphers. */
+
+  if (openpgp_cipher_get_algo_blklen (algo) != 8)
+    return 1;
+
+  if (opt.verbose)
+    warn_missing_mdc_from_pklist (pk_list);
+
+  return 0; /* No MDC */
 }
 
 
@@ -400,9 +174,9 @@ encrypt_simple (const char *filename, int mode, int use_seskey)
   PACKET pkt;
   PKT_plaintext *pt = NULL;
   STRING2KEY *s2k = NULL;
-  void *enckey = NULL;
-  size_t enckeylen = 0;
+  byte enckey[33];
   int rc = 0;
+  int seskeylen = 0;
   u32 filesize;
   cipher_filter_context_t cfx;
   armor_filter_context_t  *afx = NULL;
@@ -454,57 +228,52 @@ encrypt_simple (const char *filename, int mode, int use_seskey)
   cfx.dek = NULL;
   if ( mode )
     {
-      aead_algo_t aead_algo;
+      int canceled;
 
-      rc = setup_symkey (&s2k, &cfx.dek);
-      if (rc)
+      s2k = xmalloc_clear( sizeof *s2k );
+      s2k->mode = opt.s2k_mode;
+      s2k->hash_algo = S2K_DIGEST_ALGO;
+      cfx.dek = passphrase_to_dek (default_cipher_algo (), s2k, 1, 0,
+                                   NULL, &canceled);
+      if ( !cfx.dek || !cfx.dek->keylen )
         {
+          rc = gpg_error (canceled? GPG_ERR_CANCELED:GPG_ERR_INV_PASSPHRASE);
+          xfree (cfx.dek);
+          xfree (s2k);
           iobuf_close (inp);
-          if (gpg_err_code (rc) == GPG_ERR_CIPHER_ALGO
-              || gpg_err_code (rc) == GPG_ERR_DIGEST_ALGO)
-            ; /* Error has already been printed.  */
-          else
-            log_error (_("error creating passphrase: %s\n"), gpg_strerror (rc));
+          log_error (_("error creating passphrase: %s\n"), gpg_strerror (rc));
           release_progress_context (pfx);
           return rc;
         }
       if (use_seskey && s2k->mode != 1 && s2k->mode != 3)
         {
           use_seskey = 0;
-          log_info (_("can't use a SKESK packet due to the S2K mode\n"));
+          log_info (_("can't use a symmetric ESK packet "
+                      "due to the S2K mode\n"));
         }
 
-      /* See whether we want to use AEAD.  */
-      aead_algo = use_aead (NULL, cfx.dek->algo);
-
       if ( use_seskey )
         {
           DEK *dek = NULL;
 
-          rc = encrypt_seskey (cfx.dek, aead_algo, &dek, &enckey, &enckeylen);
-          if (rc)
-            {
-              xfree (cfx.dek);
-              xfree (s2k);
-              iobuf_close (inp);
-              release_progress_context (pfx);
-              return rc;
-            }
-          /* Replace key in DEK.  */
-          xfree (cfx.dek);
-          cfx.dek = dek;
+          seskeylen = openpgp_cipher_get_algo_keylen (default_cipher_algo ());
+          encrypt_seskey( cfx.dek, &dek, enckey );
+          xfree( cfx.dek ); cfx.dek = dek;
         }
 
-      if (aead_algo)
-        cfx.dek->use_aead = aead_algo;
-      else
-        cfx.dek->use_mdc = !!use_mdc (NULL, cfx.dek->algo);
-
       if (opt.verbose)
-        log_info(_("using cipher %s.%s\n"),
-                 openpgp_cipher_algo_name (cfx.dek->algo),
-                 cfx.dek->use_aead? openpgp_aead_algo_name (cfx.dek->use_aead)
-                 /**/             : "CFB");
+        log_info(_("using cipher %s\n"),
+                 openpgp_cipher_algo_name (cfx.dek->algo));
+
+      cfx.dek->use_mdc=use_mdc(NULL,cfx.dek->algo);
+    }
+
+  if (do_compress && cfx.dek && cfx.dek->use_mdc
+      && is_file_compressed(filename, &rc))
+    {
+      if (opt.verbose)
+        log_info(_("'%s' already compressed\n"), filename);
+      do_compress = 0;
     }
 
   if ( rc || (rc = open_outfile (-1, filename, opt.armor? 1:0, 0, &out )))
@@ -524,24 +293,20 @@ encrypt_simple (const char *filename, int mode, int use_seskey)
 
   if ( s2k )
     {
-      /* Fixme: This is quite similar to write_symkey_enc.  */
-      PKT_symkey_enc *enc = xmalloc_clear (sizeof *enc + enckeylen);
-      enc->version = cfx.dek->use_aead ? 5 : 4;
+      PKT_symkey_enc *enc = xmalloc_clear( sizeof *enc + seskeylen + 1 );
+      enc->version = 4;
       enc->cipher_algo = cfx.dek->algo;
-      enc->aead_algo = cfx.dek->use_aead;
       enc->s2k = *s2k;
-      if (enckeylen)
+      if ( use_seskey && seskeylen )
         {
-          enc->seskeylen = enckeylen;
-          memcpy (enc->seskey, enckey, enckeylen);
+          enc->seskeylen = seskeylen + 1; /* algo id */
+          memcpy (enc->seskey, enckey, seskeylen + 1 );
         }
       pkt.pkttype = PKT_SYMKEY_ENC;
       pkt.pkt.symkey_enc = enc;
       if ((rc = build_packet( out, &pkt )))
         log_error("build symkey packet failed: %s\n", gpg_strerror (rc) );
       xfree (enc);
-      xfree (enckey);
-      enckey = NULL;
     }
 
   if (!opt.no_literal)
@@ -559,12 +324,12 @@ encrypt_simple (const char *filename, int mode, int use_seskey)
 
   if ( !iobuf_is_pipe_filename (filename) && *filename && !opt.textmode )
     {
-      uint64_t tmpsize;
+      off_t tmpsize;
+      int overflow;
 
-      tmpsize = iobuf_get_filelength(inp);
-      if (!tmpsize && opt.verbose)
+      if ( !(tmpsize = iobuf_get_filelength(inp, &overflow))
+           && !overflow && opt.verbose)
         log_info(_("WARNING: '%s' is an empty file\n"), filename );
-
       /* We can't encode the length of very large files because
          OpenPGP uses only 32 bit for file sizes.  So if the
          size of a file is larger than 2^32 minus some bytes for
@@ -577,24 +342,6 @@ encrypt_simple (const char *filename, int mode, int use_seskey)
   else
     filesize = opt.set_filesize ? opt.set_filesize : 0; /* stdin */
 
-  /* Register the cipher filter. */
-  if (mode)
-    iobuf_push_filter (out,
-                       cfx.dek->use_aead? cipher_filter_aead
-                       /**/             : cipher_filter_cfb,
-                       &cfx );
-
-  if (do_compress
-      && cfx.dek
-      && (cfx.dek->use_mdc || cfx.dek->use_aead)
-      && !opt.explicit_compress_option
-      && is_file_compressed (inp))
-    {
-      if (opt.verbose)
-        log_info(_("'%s' already compressed\n"), filename? filename: "[stdin]");
-      do_compress = 0;
-    }
-
   if (!opt.no_literal)
     {
       /* Note that PT has been initialized above in !no_literal mode.  */
@@ -614,10 +361,14 @@ encrypt_simple (const char *filename, int mode, int use_seskey)
       pkt.pkt.generic = NULL;
     }
 
+  /* Register the cipher filter. */
+  if (mode)
+    iobuf_push_filter ( out, cipher_filter, &cfx );
+
   /* Register the compress filter. */
   if ( do_compress )
     {
-      if (cfx.dek && (cfx.dek->use_mdc || cfx.dek->use_aead))
+      if (cfx.dek && cfx.dek->use_mdc)
         zfx.new_ctb = 1;
       push_compress_filter (out, &zfx, default_compress_algo());
     }
@@ -632,13 +383,15 @@ encrypt_simple (const char *filename, int mode, int use_seskey)
     {
       /* User requested not to create a literal packet, so we copy the
          plain data.  */
-      iobuf_copy (out, inp);
-      if ((rc = iobuf_error (inp)))
-        log_error (_("error reading '%s': %s\n"),
-                   iobuf_get_fname_nonnull (inp), gpg_strerror (rc));
-      else if ((rc = iobuf_error (out)))
-        log_error (_("error writing '%s': %s\n"),
-                   iobuf_get_fname_nonnull (out), gpg_strerror (rc));
+    byte copy_buffer[4096];
+    int  bytes_copied;
+    while ((bytes_copied = iobuf_read(inp, copy_buffer, 4096)) != -1)
+      if ( (rc=iobuf_write(out, copy_buffer, bytes_copied)) ) {
+        log_error ("copying input to output failed: %s\n",
+                   gpg_strerror (rc) );
+        break;
+      }
+    wipememory (copy_buffer, 4096); /* burn buffer */
     }
 
   /* Finish the stuff.  */
@@ -654,7 +407,6 @@ encrypt_simple (const char *filename, int mode, int use_seskey)
   if (pt)
     pt->buf = NULL;
   free_packet (&pkt, NULL);
-  xfree (enckey);
   xfree (cfx.dek);
   xfree (s2k);
   release_armor_context (afx);
@@ -663,54 +415,22 @@ encrypt_simple (const char *filename, int mode, int use_seskey)
 }
 
 
-gpg_error_t
-setup_symkey (STRING2KEY **symkey_s2k, DEK **symkey_dek)
+int
+setup_symkey (STRING2KEY **symkey_s2k,DEK **symkey_dek)
 {
   int canceled;
-  int defcipher;
-  int s2kdigest;
 
-  defcipher = default_cipher_algo ();
-  if (openpgp_cipher_blocklen (defcipher) < 16
-      && !opt.flags.allow_old_cipher_algos)
-    {
-      log_error (_("cipher algorithm '%s' may not be used for encryption\n"),
-		 openpgp_cipher_algo_name (defcipher));
-      if (!opt.quiet)
-        log_info (_("(use option \"%s\" to override)\n"),
-                  "--allow-old-cipher-algos");
-      return gpg_error (GPG_ERR_CIPHER_ALGO);
-    }
-
-  if (!gnupg_cipher_is_allowed (opt.compliance, 1, defcipher,
-                                GCRY_CIPHER_MODE_CFB))
-    {
-      log_error (_("cipher algorithm '%s' may not be used in %s mode\n"),
-		 openpgp_cipher_algo_name (defcipher),
-		 gnupg_compliance_option_string (opt.compliance));
-      return gpg_error (GPG_ERR_CIPHER_ALGO);
-    }
-
-  s2kdigest = S2K_DIGEST_ALGO;
-  if (!gnupg_digest_is_allowed (opt.compliance, 1, s2kdigest))
-    {
-      log_error (_("digest algorithm '%s' may not be used in %s mode\n"),
-		 gcry_md_algo_name (s2kdigest),
-		 gnupg_compliance_option_string (opt.compliance));
-      return gpg_error (GPG_ERR_DIGEST_ALGO);
-    }
-
-  *symkey_s2k = xmalloc_clear (sizeof **symkey_s2k);
+  *symkey_s2k=xmalloc_clear(sizeof(STRING2KEY));
   (*symkey_s2k)->mode = opt.s2k_mode;
-  (*symkey_s2k)->hash_algo = s2kdigest;
+  (*symkey_s2k)->hash_algo = S2K_DIGEST_ALGO;
 
-  *symkey_dek = passphrase_to_dek (defcipher,
-                                   *symkey_s2k, 1, 0, NULL, 0, &canceled);
-  if (!*symkey_dek || !(*symkey_dek)->keylen)
+  *symkey_dek = passphrase_to_dek (opt.s2k_cipher_algo,
+                                   *symkey_s2k, 1, 0, NULL, &canceled);
+  if(!*symkey_dek || !(*symkey_dek)->keylen)
     {
       xfree(*symkey_dek);
       xfree(*symkey_s2k);
-      return gpg_error (canceled?GPG_ERR_CANCELED:GPG_ERR_INV_PASSPHRASE);
+      return gpg_error (canceled?GPG_ERR_CANCELED:GPG_ERR_BAD_PASSPHRASE);
     }
 
   return 0;
@@ -718,33 +438,23 @@ setup_symkey (STRING2KEY **symkey_s2k, DEK **symkey_dek)
 
 
 static int
-write_symkey_enc (STRING2KEY *symkey_s2k, aead_algo_t aead_algo,
-                  DEK *symkey_dek, DEK *dek, iobuf_t out)
+write_symkey_enc (STRING2KEY *symkey_s2k, DEK *symkey_dek, DEK *dek,
+                  iobuf_t out)
 {
-  int rc;
-  void *enckey;
-  size_t enckeylen;
+  int rc, seskeylen = openpgp_cipher_get_algo_keylen (dek->algo);
+
   PKT_symkey_enc *enc;
+  byte enckey[33];
   PACKET pkt;
 
-  rc = encrypt_seskey (symkey_dek, aead_algo, &dek, &enckey, &enckeylen);
-  if (rc)
-    return rc;
-  enc = xtrycalloc (1, sizeof (PKT_symkey_enc) + enckeylen);
-  if (!enc)
-    {
-      rc = gpg_error_from_syserror ();
-      xfree (enckey);
-      return rc;
-    }
+  enc=xmalloc_clear(sizeof(PKT_symkey_enc)+seskeylen+1);
+  encrypt_seskey(symkey_dek,&dek,enckey);
 
-  enc->version = aead_algo? 5 : 4;
+  enc->version = 4;
   enc->cipher_algo = opt.s2k_cipher_algo;
-  enc->aead_algo = aead_algo;
   enc->s2k = *symkey_s2k;
-  enc->seskeylen = enckeylen;
-  memcpy (enc->seskey, enckey, enckeylen);
-  xfree (enckey);
+  enc->seskeylen = seskeylen + 1; /* algo id */
+  memcpy( enc->seskey, enckey, seskeylen + 1 );
 
   pkt.pkttype = PKT_SYMKEY_ENC;
   pkt.pkt.symkey_enc = enc;
@@ -752,7 +462,7 @@ write_symkey_enc (STRING2KEY *symkey_s2k, aead_algo_t aead_algo,
   if ((rc=build_packet(out,&pkt)))
     log_error("build symkey_enc packet failed: %s\n",gpg_strerror (rc));
 
-  xfree (enc);
+  xfree(enc);
   return rc;
 }
 
@@ -763,8 +473,8 @@ write_symkey_enc (STRING2KEY *symkey_s2k, aead_algo_t aead_algo,
  * The caller may provide a checked list of public keys in
  * PROVIDED_PKS; if not the function builds a list of keys on its own.
  *
- * Note that FILEFD and OUTPUTFD are currently only used by
- * cmd_encrypt in the not yet finished server.c.
+ * Note that FILEFD is currently only used by cmd_encrypt in the
+ * not yet finished server.c.
  */
 int
 encrypt_crypt (ctrl_t ctrl, int filefd, const char *filename,
@@ -777,7 +487,7 @@ encrypt_crypt (ctrl_t ctrl, int filefd, const char *filename,
   PKT_plaintext *pt = NULL;
   DEK *symkey_dek = NULL;
   STRING2KEY *symkey_s2k = NULL;
-  int rc = 0;
+  int rc = 0, rc2 = 0;
   u32 filesize;
   cipher_filter_context_t cfx;
   armor_filter_context_t *afx = NULL;
@@ -786,6 +496,7 @@ encrypt_crypt (ctrl_t ctrl, int filefd, const char *filename,
   progress_filter_context_t *pfx;
   PK_LIST pk_list;
   int do_compress;
+  int compliant;
 
   if (filefd != -1 && filename)
     return gpg_error (GPG_ERR_INV_ARG);  /* Both given.  */
@@ -826,10 +537,10 @@ encrypt_crypt (ctrl_t ctrl, int filefd, const char *filename,
       gpg_err_set_errno (ENOSYS);
     }
 #else
-  if (filefd == -1)
+  if (filefd == GNUPG_INVALID_FD)
     inp = iobuf_open (filename);
   else
-    inp = iobuf_fdopen_nc (filefd, "rb");
+    inp = iobuf_fdopen_nc (FD2INT(filefd), "rb");
 #endif
   if (inp)
     iobuf_ioctl (inp, IOBUF_IOCTL_NO_CACHE, 1, NULL);
@@ -874,30 +585,129 @@ encrypt_crypt (ctrl_t ctrl, int filefd, const char *filename,
     }
 
   /* Create a session key. */
-  cfx.dek = create_dek_with_warnings (pk_list);
+  cfx.dek = xmalloc_secure_clear (sizeof *cfx.dek);
+  if (!opt.def_cipher_algo)
+    {
+      /* Try to get it from the prefs.  */
+      cfx.dek->algo = select_algo_from_prefs (pk_list, PREFTYPE_SYM, -1, NULL);
+      /* The only way select_algo_from_prefs can fail here is when
+         mixing v3 and v4 keys, as v4 keys have an implicit preference
+         entry for 3DES, and the pk_list cannot be empty.  In this
+         case, use 3DES anyway as it's the safest choice - perhaps the
+         v3 key is being used in an OpenPGP implementation and we know
+         that the implementation behind any v4 key can handle 3DES. */
+      if (cfx.dek->algo == -1)
+        {
+          cfx.dek->algo = CIPHER_ALGO_3DES;
+        }
 
-  rc = check_encryption_compliance (cfx.dek, pk_list);
-  if (rc)
-    goto leave;
+      /* In case 3DES has been selected, print a warning if any key
+         does not have a preference for AES.  This should help to
+         indentify why encrypting to several recipients falls back to
+         3DES. */
+      if (opt.verbose && cfx.dek->algo == CIPHER_ALGO_3DES)
+        warn_missing_aes_from_pklist (pk_list);
+    }
+  else
+    {
+      if (!opt.expert
+          && (select_algo_from_prefs (pk_list, PREFTYPE_SYM,
+                                      opt.def_cipher_algo, NULL)
+              != opt.def_cipher_algo))
+        {
+          log_info(_("WARNING: forcing symmetric cipher %s (%d)"
+                     " violates recipient preferences\n"),
+                   openpgp_cipher_algo_name (opt.def_cipher_algo),
+                   opt.def_cipher_algo);
+        }
 
-  cfx.dek->use_aead = use_aead (pk_list, cfx.dek->algo);
-  if (!cfx.dek->use_aead)
-    cfx.dek->use_mdc = !!use_mdc (pk_list, cfx.dek->algo);
+      cfx.dek->algo = opt.def_cipher_algo;
+    }
+
+  /* Check compliance.  */
+  if (! gnupg_cipher_is_allowed (opt.compliance, 1, cfx.dek->algo,
+                                 GCRY_CIPHER_MODE_CFB))
+    {
+      log_error (_("cipher algorithm '%s' may not be used in %s mode\n"),
+		 openpgp_cipher_algo_name (cfx.dek->algo),
+		 gnupg_compliance_option_string (opt.compliance));
+      rc = gpg_error (GPG_ERR_CIPHER_ALGO);
+      goto leave;
+    }
+
+  if (!gnupg_rng_is_compliant (opt.compliance))
+    {
+      rc = gpg_error (GPG_ERR_FORBIDDEN);
+      log_error (_("%s is not compliant with %s mode\n"),
+                 "RNG",
+                 gnupg_compliance_option_string (opt.compliance));
+      write_status_error ("random-compliance", rc);
+      goto leave;
+    }
+
+  compliant = gnupg_cipher_is_compliant (CO_DE_VS, cfx.dek->algo,
+                                         GCRY_CIPHER_MODE_CFB);
+
+  {
+    pk_list_t pkr;
+
+    for (pkr = pk_list; pkr; pkr = pkr->next)
+      {
+        PKT_public_key *pk = pkr->pk;
+        unsigned int nbits = nbits_from_pk (pk);
+
+        if (!gnupg_pk_is_compliant (opt.compliance,
+                                    pk->pubkey_algo, pk->pkey, nbits, NULL))
+          log_info (_("WARNING: key %s is not suitable for encryption"
+                      " in %s mode\n"),
+                    keystr_from_pk (pk),
+                    gnupg_compliance_option_string (opt.compliance));
+
+        if (compliant
+            && !gnupg_pk_is_compliant (CO_DE_VS, pk->pubkey_algo, pk->pkey,
+                                       nbits, NULL))
+          compliant = 0;
+      }
+
+  }
+
+  if (compliant)
+    write_status_strings (STATUS_ENCRYPTION_COMPLIANCE_MODE,
+                          gnupg_status_compliance_flag (CO_DE_VS),
+                          NULL);
+
+  cfx.dek->use_mdc = use_mdc (pk_list,cfx.dek->algo);
+
+  /* Only do the is-file-already-compressed check if we are using a
+     MDC.  This forces compressed files to be re-compressed if we do
+     not have a MDC to give some protection against chosen ciphertext
+     attacks. */
+
+  if (do_compress && cfx.dek->use_mdc && is_file_compressed(filename, &rc2))
+    {
+      if (opt.verbose)
+        log_info(_("'%s' already compressed\n"), filename);
+      do_compress = 0;
+    }
+  if (rc2)
+    {
+      rc = rc2;
+      goto leave;
+    }
 
   make_session_key (cfx.dek);
   if (DBG_CRYPTO)
-    log_printhex (cfx.dek->key, cfx.dek->keylen, "DEK is: ");
+    log_printhex ("DEK is: ", cfx.dek->key, cfx.dek->keylen );
 
   rc = write_pubkey_enc_from_list (ctrl, pk_list, cfx.dek, out);
   if (rc)
     goto leave;
 
   /* We put the passphrase (if any) after any public keys as this
-   * seems to be the most useful on the recipient side - there is no
-   * point in prompting a user for a passphrase if they have the
-   * secret key needed to decrypt.  */
-  if (use_symkey && (rc = write_symkey_enc (symkey_s2k, cfx.dek->use_aead,
-                                            symkey_dek, cfx.dek, out)))
+     seems to be the most useful on the recipient side - there is no
+     point in prompting a user for a passphrase if they have the
+     secret key needed to decrypt.  */
+  if(use_symkey && (rc = write_symkey_enc(symkey_s2k,symkey_dek,cfx.dek,out)))
     goto leave;
 
   if (!opt.no_literal)
@@ -907,10 +717,11 @@ encrypt_crypt (ctrl_t ctrl, int filefd, const char *filename,
   if (filename && *filename
       && !iobuf_is_pipe_filename (filename) && !opt.textmode )
     {
-      uint64_t tmpsize;
+      off_t tmpsize;
+      int overflow;
 
-      tmpsize = iobuf_get_filelength (inp);
-      if (!tmpsize && opt.verbose)
+      if ( !(tmpsize = iobuf_get_filelength(inp, &overflow))
+           && !overflow && opt.verbose)
         log_info(_("WARNING: '%s' is an empty file\n"), filename );
       /* We can't encode the length of very large files because
          OpenPGP uses only 32 bit for file sizes.  So if the size
@@ -924,26 +735,6 @@ encrypt_crypt (ctrl_t ctrl, int filefd, const char *filename,
   else
     filesize = opt.set_filesize ? opt.set_filesize : 0; /* stdin */
 
-  /* Register the cipher filter. */
-  iobuf_push_filter (out,
-                     cfx.dek->use_aead? cipher_filter_aead
-                     /**/             : cipher_filter_cfb,
-                     &cfx);
-
-  /* Only do the is-file-already-compressed check if we are using a
-   * MDC or AEAD.  This forces compressed files to be re-compressed if
-   * we do not have a MDC to give some protection against chosen
-   * ciphertext attacks. */
-  if (do_compress
-      && (cfx.dek->use_mdc || cfx.dek->use_aead)
-      && !opt.explicit_compress_option
-      && is_file_compressed (inp))
-    {
-      if (opt.verbose)
-        log_info(_("'%s' already compressed\n"), filename? filename: "[stdin]");
-      do_compress = 0;
-    }
-
   if (!opt.no_literal)
     {
       pt->timestamp = make_timestamp();
@@ -958,6 +749,9 @@ encrypt_crypt (ctrl_t ctrl, int filefd, const char *filename,
   else
     cfx.datalen = filesize && !do_compress ? filesize : 0;
 
+  /* Register the cipher filter. */
+  iobuf_push_filter (out, cipher_filter, &cfx);
+
   /* Register the compress filter. */
   if (do_compress)
     {
@@ -983,7 +777,7 @@ encrypt_crypt (ctrl_t ctrl, int filefd, const char *filename,
       /* Algo 0 means no compression. */
       if (compr_algo)
         {
-          if (cfx.dek && (cfx.dek->use_mdc || cfx.dek->use_aead))
+          if (cfx.dek && cfx.dek->use_mdc)
             zfx.new_ctb = 1;
           push_compress_filter (out,&zfx,compr_algo);
         }
@@ -999,14 +793,19 @@ encrypt_crypt (ctrl_t ctrl, int filefd, const char *filename,
     {
       /* User requested not to create a literal packet, so we copy the
          plain data. */
-      iobuf_copy (out, inp);
-      if ((rc = iobuf_error (inp)))
-        log_error (_("error reading '%s': %s\n"),
-                   iobuf_get_fname_nonnull (inp), gpg_strerror (rc));
-      else if ((rc = iobuf_error (out)))
-        log_error (_("error writing '%s': %s\n"),
-                   iobuf_get_fname_nonnull (out), gpg_strerror (rc));
-
+      byte copy_buffer[4096];
+      int  bytes_copied;
+      while ((bytes_copied = iobuf_read (inp, copy_buffer, 4096)) != -1)
+        {
+          rc = iobuf_write (out, copy_buffer, bytes_copied);
+          if (rc)
+            {
+              log_error ("copying input to output failed: %s\n",
+                         gpg_strerror (rc));
+              break;
+            }
+        }
+      wipememory (copy_buffer, 4096); /* Burn the buffer. */
     }
 
   /* Finish the stuff. */
@@ -1052,21 +851,47 @@ encrypt_filter (void *opaque, int control,
     {
       if ( !efx->header_okay )
         {
-          efx->header_okay = 1;
+          efx->cfx.dek = xmalloc_secure_clear ( sizeof *efx->cfx.dek );
+          if ( !opt.def_cipher_algo  )
+            {
+              /* Try to get it from the prefs. */
+              efx->cfx.dek->algo =
+                select_algo_from_prefs (efx->pk_list, PREFTYPE_SYM, -1, NULL);
+              if (efx->cfx.dek->algo == -1 )
+                {
+                  /* Because 3DES is implicitly in the prefs, this can
+                     only happen if we do not have any public keys in
+                     the list.  */
+                  efx->cfx.dek->algo = DEFAULT_CIPHER_ALGO;
+                }
 
-          efx->cfx.dek = create_dek_with_warnings (efx->pk_list);
+              /* In case 3DES has been selected, print a warning if
+                 any key does not have a preference for AES.  This
+                 should help to indentify why encrypting to several
+                 recipients falls back to 3DES. */
+              if (opt.verbose
+                  && efx->cfx.dek->algo == CIPHER_ALGO_3DES)
+                warn_missing_aes_from_pklist (efx->pk_list);
+	    }
+          else
+            {
+	      if (!opt.expert
+                  && select_algo_from_prefs (efx->pk_list,PREFTYPE_SYM,
+                                             opt.def_cipher_algo,
+                                             NULL) != opt.def_cipher_algo)
+		log_info(_("forcing symmetric cipher %s (%d) "
+			   "violates recipient preferences\n"),
+			 openpgp_cipher_algo_name (opt.def_cipher_algo),
+			 opt.def_cipher_algo);
 
-          rc = check_encryption_compliance (efx->cfx.dek, efx->pk_list);
-          if (rc)
-            return rc;
+	      efx->cfx.dek->algo = opt.def_cipher_algo;
+	    }
 
-          efx->cfx.dek->use_aead = use_aead (efx->pk_list, efx->cfx.dek->algo);
-          if (!efx->cfx.dek->use_aead)
-            efx->cfx.dek->use_mdc = !!use_mdc (efx->pk_list,efx->cfx.dek->algo);
+          efx->cfx.dek->use_mdc = use_mdc (efx->pk_list,efx->cfx.dek->algo);
 
           make_session_key ( efx->cfx.dek );
           if (DBG_CRYPTO)
-            log_printhex (efx->cfx.dek->key, efx->cfx.dek->keylen, "DEK is: ");
+            log_printhex ("DEK is: ", efx->cfx.dek->key, efx->cfx.dek->keylen);
 
           rc = write_pubkey_enc_from_list (efx->ctrl,
                                            efx->pk_list, efx->cfx.dek, a);
@@ -1075,17 +900,15 @@ encrypt_filter (void *opaque, int control,
 
           if(efx->symkey_s2k && efx->symkey_dek)
             {
-              rc = write_symkey_enc (efx->symkey_s2k, efx->cfx.dek->use_aead,
-                                     efx->symkey_dek, efx->cfx.dek, a);
-              if (rc)
+              rc=write_symkey_enc(efx->symkey_s2k,efx->symkey_dek,
+                                  efx->cfx.dek,a);
+              if(rc)
                 return rc;
             }
 
-          iobuf_push_filter (a,
-                             efx->cfx.dek->use_aead? cipher_filter_aead
-                             /**/                  : cipher_filter_cfb,
-                             &efx->cfx);
+          iobuf_push_filter (a, cipher_filter, &efx->cfx);
 
+          efx->header_okay = 1;
         }
       rc = iobuf_write (a, buf, size);
 
@@ -1144,17 +967,9 @@ write_pubkey_enc (ctrl_t ctrl,
       if ( opt.verbose )
         {
           char *ustr = get_user_id_string_native (ctrl, enc->keyid);
-          if ((pk->pubkey_usage & PUBKEY_USAGE_RENC))
-            {
-              char *tmpustr = xstrconcat (ustr, " [ADSK]", NULL);
-              xfree (ustr);
-              ustr = tmpustr;
-            }
-          log_info (_("%s/%s.%s encrypted for: \"%s\"\n"),
+          log_info (_("%s/%s encrypted for: \"%s\"\n"),
                     openpgp_pk_algo_name (enc->pubkey_algo),
                     openpgp_cipher_algo_name (dek->algo),
-                    dek->use_aead? openpgp_aead_algo_name (dek->use_aead)
-                    /**/         : "CFB",
                     ustr );
           xfree (ustr);
         }
@@ -1178,7 +993,7 @@ write_pubkey_enc (ctrl_t ctrl,
 static int
 write_pubkey_enc_from_list (ctrl_t ctrl, PK_LIST pk_list, DEK *dek, iobuf_t out)
 {
-  if (opt.throw_keyids && (PGP7 || PGP8))
+  if (opt.throw_keyids && (PGP6 || PGP7 || PGP8))
     {
       log_info(_("option '%s' may not be used in %s mode\n"),
                "--throw-keyids",
