@@ -1,7 +1,6 @@
 /* gpgv.c - The GnuPG signature verify utility
- * Copyright (C) 1998-2020 Free Software Foundation, Inc.
- * Copyright (C) 1997-2019 Werner Koch
- * Copyright (C) 2015-2020 g10 Code GmbH
+ * Copyright (C) 1998, 1999, 2000, 2001, 2002, 2005, 2006,
+ *               2008, 2009, 2012 Free Software Foundation, Inc.
  *
  * This file is part of GnuPG.
  *
@@ -17,7 +16,6 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, see <https://www.gnu.org/licenses/>.
- * SPDX-License-Identifier: GPL-3.0-or-later
  */
 
 #include <config.h>
@@ -68,12 +66,11 @@ enum cmd_and_opt_values {
   oWeakDigest,
   oEnableSpecialFilenames,
   oDebug,
-  oAssertPubkeyAlgo,
   aTest
 };
 
 
-static gpgrt_opt_t opts[] = {
+static ARGPARSE_OPTS opts[] = {
   ARGPARSE_group (300, N_("@\nOptions:\n ")),
 
   ARGPARSE_s_n (oVerbose, "verbose", N_("verbose")),
@@ -92,7 +89,6 @@ static gpgrt_opt_t opts[] = {
                 N_("|ALGO|reject signatures made with ALGO")),
   ARGPARSE_s_n (oEnableSpecialFilenames, "enable-special-filenames", "@"),
   ARGPARSE_s_s (oDebug, "debug", "@"),
-  ARGPARSE_s_s (oAssertPubkeyAlgo,"assert-pubkey-algo", "@"),
 
   ARGPARSE_end ()
 };
@@ -120,8 +116,7 @@ static struct debug_flags_s debug_flags [] =
 
 
 int g10_errors_seen = 0;
-int assert_signer_true = 0;
-int assert_pubkey_algo_false = 0;
+
 
 static char *
 make_libversion (const char *libname, const char *(*getfnc)(const char*))
@@ -143,11 +138,9 @@ my_strusage( int level )
 
   switch (level)
     {
-    case  9: p = "GPL-3.0-or-later"; break;
     case 11: p = "@GPG@v (GnuPG)";
       break;
     case 13: p = VERSION; break;
-    case 14: p = GNUPG_DEF_COPYRIGHT_LINE; break;
     case 17: p = PRINTABLE_OS_NAME; break;
     case 19: p = _("Please report bugs to <@EMAIL@>.\n"); break;
 
@@ -175,14 +168,15 @@ my_strusage( int level )
 int
 main( int argc, char **argv )
 {
-  gpgrt_argparse_t pargs;
+  ARGPARSE_ARGS pargs;
   int rc=0;
   strlist_t sl;
   strlist_t nrings = NULL;
+  unsigned configlineno;
   ctrl_t ctrl;
 
   early_system_init ();
-  gpgrt_set_strusage (my_strusage);
+  set_strusage (my_strusage);
   log_set_prefix ("gpgv", GPGRT_LOG_WITH_PREFIX);
 
   /* Make sure that our subsystems are ready.  */
@@ -212,13 +206,11 @@ main( int argc, char **argv )
 
   pargs.argc = &argc;
   pargs.argv = &argv;
-  pargs.flags= ARGPARSE_FLAG_KEEP;
-  while (gpgrt_argparser (&pargs, opts, NULL))
+  pargs.flags=  1;  /* do not remove the args */
+  while (optfile_parse( NULL, NULL, &configlineno, &pargs, opts))
     {
       switch (pargs.r_opt)
         {
-        case ARGPARSE_CONFFILE: break;
-
         case oQuiet: opt.quiet = 1; break;
         case oVerbose:
           opt.verbose++;
@@ -254,25 +246,10 @@ main( int argc, char **argv )
         case oEnableSpecialFilenames:
           enable_special_filenames ();
           break;
-
-        case oAssertPubkeyAlgo:
-          if (!opt.assert_pubkey_algos)
-            opt.assert_pubkey_algos = xstrdup (pargs.r.ret_str);
-          else
-            {
-              char *tmp = opt.assert_pubkey_algos;
-              opt.assert_pubkey_algos = xstrconcat (tmp, ",",
-                                                    pargs.r.ret_str, NULL);
-              xfree (tmp);
-            }
-          break;
-
         default : pargs.err = ARGPARSE_PRINT_ERROR; break;
 	}
     }
 
-  gpgrt_argparse (NULL, &pargs, NULL);  /* Release internal state.  */
-
   if (log_get_errorcount (0))
     g10_exit(2);
 
@@ -304,18 +281,10 @@ main( int argc, char **argv )
 
 
 void
-g10_exit (int rc)
+g10_exit( int rc )
 {
-  if (rc)
-    ;
-  else if (log_get_errorcount(0))
-    rc = 2;
-  else if (g10_errors_seen)
-    rc = 1;
-  else if (opt.assert_pubkey_algos && assert_pubkey_algo_false)
-    rc = 1;
-
-  exit (rc);
+  rc = rc? rc : log_get_errorcount(0)? 2 : g10_errors_seen? 1 : 0;
+  exit(rc );
 }
 
 
@@ -323,13 +292,10 @@ g10_exit (int rc)
  * We have to override the trustcheck from pkclist.c because
  * this utility assumes that all keys in the keyring are trustworthy
  */
-gpg_error_t
-check_signatures_trust (ctrl_t ctrl, kbnode_t kblock,
-                        PKT_public_key *pk, PKT_signature *sig)
+int
+check_signatures_trust (ctrl_t ctrl, PKT_signature *sig)
 {
   (void)ctrl;
-  (void)kblock;
-  (void)pk;
   (void)sig;
   return 0;
 }
@@ -426,14 +392,6 @@ get_ownertrust (ctrl_t ctrl, PKT_public_key *pk)
   return TRUST_UNKNOWN;
 }
 
-const char *
-get_ownertrust_string (ctrl_t ctrl, PKT_public_key *pk, int no_create)
-{
-  (void)ctrl;
-  (void)pk;
-  (void)no_create;
-  return "";
-}
 
 /* Stubs:
  * Because we only work with trusted keys, it does not make sense to
@@ -455,33 +413,23 @@ keyserver_any_configured (ctrl_t ctrl)
 }
 
 int
-keyserver_import_keyid (u32 *keyid, void *dummy, unsigned int flags)
+keyserver_import_keyid (u32 *keyid, void *dummy, int quick)
 {
   (void)keyid;
   (void)dummy;
-  (void)flags;
+  (void)quick;
   return -1;
 }
 
 int
 keyserver_import_fprint (ctrl_t ctrl, const byte *fprint,size_t fprint_len,
-			 struct keyserver_spec *keyserver, unsigned int flags)
+			 struct keyserver_spec *keyserver, int quick)
 {
   (void)ctrl;
   (void)fprint;
   (void)fprint_len;
   (void)keyserver;
-  (void)flags;
-  return -1;
-}
-
-int
-keyserver_import_fprint_ntds (ctrl_t ctrl,
-                              const byte *fprint, size_t fprint_len)
-{
-  (void)ctrl;
-  (void)fprint;
-  (void)fprint_len;
+  (void)quick;
   return -1;
 }
 
@@ -492,37 +440,34 @@ keyserver_import_cert (const char *name)
   return -1;
 }
 
+int
+keyserver_import_pka (const char *name,unsigned char *fpr)
+{
+  (void)name;
+  (void)fpr;
+  return -1;
+}
+
 gpg_error_t
-keyserver_import_wkd (ctrl_t ctrl, const char *name, unsigned int flags,
+keyserver_import_wkd (ctrl_t ctrl, const char *name, int quick,
                       unsigned char **fpr, size_t *fpr_len)
 {
   (void)ctrl;
   (void)name;
-  (void)flags;
+  (void)quick;
   (void)fpr;
   (void)fpr_len;
   return GPG_ERR_BUG;
 }
 
 int
-keyserver_import_mbox (const char *name,struct keyserver_spec *spec)
+keyserver_import_name (const char *name,struct keyserver_spec *spec)
 {
   (void)name;
   (void)spec;
   return -1;
 }
 
-int
-keyserver_import_ntds (ctrl_t ctrl, const char *mbox,
-                       unsigned char **fpr, size_t *fprlen)
-{
-  (void)ctrl;
-  (void)mbox;
-  (void)fpr;
-  (void)fprlen;
-  return -1;
-}
-
 int
 keyserver_import_ldap (const char *name)
 {
@@ -532,32 +477,20 @@ keyserver_import_ldap (const char *name)
 
 
 gpg_error_t
-read_key_from_file_or_buffer (ctrl_t ctrl, const char *fname,
-                              const void *buffer, size_t buflen,
-                              kbnode_t *r_keyblock)
+read_key_from_file (ctrl_t ctrl, const char *fname, kbnode_t *r_keyblock)
 {
   (void)ctrl;
   (void)fname;
-  (void)buffer;
-  (void)buflen;
   (void)r_keyblock;
   return -1;
 }
 
-gpg_error_t
-import_included_key_block (ctrl_t ctrl, kbnode_t keyblock)
-{
-  (void)ctrl;
-  (void)keyblock;
-  return -1;
-}
-
 
 /* Stub:
  * No encryption here but mainproc links to these functions.
  */
 gpg_error_t
-get_session_key (ctrl_t ctrl, struct pubkey_enc_list *k, DEK *dek)
+get_session_key (ctrl_t ctrl, PKT_pubkey_enc *k, DEK *dek)
 {
   (void)ctrl;
   (void)k;
@@ -576,14 +509,12 @@ get_override_session_key (DEK *dek, const char *string)
 
 /* Stub: */
 int
-decrypt_data (ctrl_t ctrl, void *procctx, PKT_encrypted *ed, DEK *dek,
-              int *compliance_error)
+decrypt_data (ctrl_t ctrl, void *procctx, PKT_encrypted *ed, DEK *dek)
 {
   (void)ctrl;
   (void)procctx;
   (void)ed;
   (void)dek;
-  (void)compliance_error;
   return GPG_ERR_GENERAL;
 }
 
@@ -613,14 +544,13 @@ check_secret_key (PKT_public_key *pk, int n)
  */
 DEK *
 passphrase_to_dek (int cipher_algo, STRING2KEY *s2k, int create, int nocache,
-                   const char *tmp, unsigned int flags, int *canceled)
+                   const char *tmp, int *canceled)
 {
   (void)cipher_algo;
   (void)s2k;
   (void)create;
   (void)nocache;
   (void)tmp;
-  (void)flags;
 
   if (canceled)
     *canceled = 0;
@@ -733,12 +663,12 @@ dotlock_remove_lockfiles (void)
 {
 }
 
-int
+gpg_error_t
 agent_probe_secret_key (ctrl_t ctrl, PKT_public_key *pk)
 {
   (void)ctrl;
   (void)pk;
-  return 0;
+  return gpg_error (GPG_ERR_NO_SECKEY);
 }
 
 gpg_error_t
@@ -760,17 +690,30 @@ agent_get_keyinfo (ctrl_t ctrl, const char *hexkeygrip,
   return gpg_error (GPG_ERR_NO_SECKEY);
 }
 
+gpg_error_t
+gpg_dirmngr_get_pka (ctrl_t ctrl, const char *userid,
+                     unsigned char **r_fpr, size_t *r_fprlen,
+                     char **r_url)
+{
+  (void)ctrl;
+  (void)userid;
+  if (r_fpr)
+    *r_fpr = NULL;
+  if (r_fprlen)
+    *r_fprlen = 0;
+  if (r_url)
+    *r_url = NULL;
+  return gpg_error (GPG_ERR_NOT_FOUND);
+}
+
 gpg_error_t
 export_pubkey_buffer (ctrl_t ctrl, const char *keyspec, unsigned int options,
-                      const void *prefix, size_t prefixlen,
                       export_stats_t stats,
                       kbnode_t *r_keyblock, void **r_data, size_t *r_datalen)
 {
   (void)ctrl;
   (void)keyspec;
   (void)options;
-  (void)prefix;
-  (void)prefixlen;
   (void)stats;
 
   *r_keyblock = NULL;
@@ -829,26 +772,3 @@ tofu_notice_key_changed (ctrl_t ctrl, kbnode_t kb)
 
   return 0;
 }
-
-
-int
-get_revocation_reason (PKT_signature *sig, char **r_reason,
-                       char **r_comment, size_t *r_commentlen)
-{
-  (void)sig;
-  (void)r_commentlen;
-
-  if (r_reason)
-    *r_reason = NULL;
-  if (r_comment)
-    *r_comment = NULL;
-  return 0;
-}
-
-const char *
-impex_filter_getval (void *cookie, const char *propname)
-{
-  (void)cookie;
-  (void)propname;
-  return NULL;
-}
