@@ -1,7 +1,6 @@
 /* scdaemon.c  -  The GnuPG Smartcard Daemon
- * Copyright (C) 2001-2002, 2004-2005, 2007-2020 Free Software Foundation, Inc.
- * Copyright (C) 2001-2002, 2004-2005, 2007-2019 Werner Koch
- * Copyright (C) 2020 g10 Code GmbH
+ * Copyright (C) 2001-2002, 2004-2005, 2007-2009 Free Software Foundation, Inc.
+ * Copyright (C) 2001-2002, 2004-2005, 2007-2014 Werner Koch
  *
  * This file is part of GnuPG.
  *
@@ -17,7 +16,6 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, see <https://www.gnu.org/licenses/>.
- * SPDX-License-Identifier: GPL-3.0-or-later
  */
 
 #include <config.h>
@@ -28,6 +26,7 @@
 #include <stdarg.h>
 #include <string.h>
 #include <errno.h>
+#include <assert.h>
 #include <time.h>
 #include <fcntl.h>
 #ifndef HAVE_W32_SYSTEM
@@ -38,7 +37,6 @@
 #include <signal.h>
 #include <npth.h>
 
-#define INCLUDED_BY_MAIN_MODULE 1
 #define GNUPG_COMMON_NEED_AFLOCAL
 #include "scdaemon.h"
 #include <ksba.h>
@@ -48,13 +46,13 @@
 
 #include "../common/i18n.h"
 #include "../common/sysutils.h"
+#include "app-common.h"
 #include "iso7816.h"
 #include "apdu.h"
 #include "ccid-driver.h"
 #include "../common/gc-opt-flags.h"
 #include "../common/asshelp.h"
 #include "../common/exechelp.h"
-#include "../common/comopt.h"
 #include "../common/init.h"
 
 #ifndef ENAMETOOLONG
@@ -79,7 +77,6 @@ enum cmd_and_opt_values
   oDebugAllowCoreDump,
   oDebugCCIDDriver,
   oDebugLogTid,
-  oDebugAllowPINLogging,
   oDebugAssuanLogCats,
   oNoGreeting,
   oNoOptions,
@@ -95,43 +92,33 @@ enum cmd_and_opt_values
   oCardTimeout,
   octapiDriver,
   opcscDriver,
-  opcscShared,
   oDisableCCID,
   oDisableOpenSC,
   oDisablePinpad,
   oAllowAdmin,
   oDenyAdmin,
   oDisableApplication,
-  oApplicationPriority,
   oEnablePinpadVarlen,
-  oCompatibilityFlags,
   oListenBacklog
 };
 
 
 
-static gpgrt_opt_t opts[] = {
+static ARGPARSE_OPTS opts[] = {
   ARGPARSE_c (aGPGConfList, "gpgconf-list", "@"),
   ARGPARSE_c (aGPGConfTest, "gpgconf-test", "@"),
 
-  ARGPARSE_header (NULL, N_("Options used for startup")),
+  ARGPARSE_group (301, N_("@Options:\n ")),
 
   ARGPARSE_s_n (oServer,"server", N_("run in server mode (foreground)")),
   ARGPARSE_s_n (oMultiServer, "multi-server",
                 N_("run in multi server mode (foreground)")),
   ARGPARSE_s_n (oDaemon, "daemon", N_("run in daemon mode (background)")),
-  ARGPARSE_s_n (oNoDetach, "no-detach", N_("do not detach from the console")),
-  ARGPARSE_s_n (oSh,    "sh", N_("sh-style command output")),
-  ARGPARSE_s_n (oCsh,   "csh", N_("csh-style command output")),
-  ARGPARSE_s_s (oHomedir,    "homedir",      "@"),
-  ARGPARSE_conffile (oOptions, "options", N_("|FILE|read options from FILE")),
-  ARGPARSE_noconffile (oNoOptions, "no-options", "@"),
-
-
-  ARGPARSE_header ("Monitor", N_("Options controlling the diagnostic output")),
-
   ARGPARSE_s_n (oVerbose, "verbose", N_("verbose")),
   ARGPARSE_s_n (oQuiet, "quiet", N_("be somewhat more quiet")),
+  ARGPARSE_s_n (oSh,    "sh", N_("sh-style command output")),
+  ARGPARSE_s_n (oCsh,   "csh", N_("csh-style command output")),
+  ARGPARSE_s_s (oOptions, "options", N_("|FILE|read options from FILE")),
   ARGPARSE_s_s (oDebug, "debug", "@"),
   ARGPARSE_s_n (oDebugAll, "debug-all", "@"),
   ARGPARSE_s_s (oDebugLevel, "debug-level" ,
@@ -140,21 +127,15 @@ static gpgrt_opt_t opts[] = {
   ARGPARSE_s_n (oDebugAllowCoreDump, "debug-allow-core-dump", "@"),
   ARGPARSE_s_n (oDebugCCIDDriver, "debug-ccid-driver", "@"),
   ARGPARSE_s_n (oDebugLogTid, "debug-log-tid", "@"),
-  ARGPARSE_s_n (oDebugAllowPINLogging, "debug-allow-pin-logging", "@"),
   ARGPARSE_p_u (oDebugAssuanLogCats, "debug-assuan-log-cats", "@"),
+  ARGPARSE_s_n (oNoDetach, "no-detach", N_("do not detach from the console")),
   ARGPARSE_s_s (oLogFile,  "log-file", N_("|FILE|write a log to FILE")),
-
-
-  ARGPARSE_header ("Configuration",
-                   N_("Options controlling the configuration")),
-
   ARGPARSE_s_s (oReaderPort, "reader-port",
                 N_("|N|connect to reader at port N")),
   ARGPARSE_s_s (octapiDriver, "ctapi-driver",
                 N_("|NAME|use NAME as ct-API driver")),
   ARGPARSE_s_s (opcscDriver, "pcsc-driver",
                 N_("|NAME|use NAME as PC/SC driver")),
-  ARGPARSE_s_n (opcscShared, "pcsc-shared", "@"),
   ARGPARSE_s_n (oDisableCCID, "disable-ccid",
 #ifdef HAVE_LIBUSB
                                     N_("do not use the internal CCID driver")
@@ -168,21 +149,15 @@ static gpgrt_opt_t opts[] = {
   ARGPARSE_s_n (oDisablePinpad, "disable-pinpad",
                 N_("do not use a reader's pinpad")),
   ARGPARSE_ignore (300, "disable-keypad"),
-  ARGPARSE_s_n (oEnablePinpadVarlen, "enable-pinpad-varlen",
-                N_("use variable length input for pinpad")),
-  ARGPARSE_s_s (oDisableApplication, "disable-application", "@"),
-  ARGPARSE_s_s (oApplicationPriority, "application-priority",
-                N_("|LIST|change the application priority to LIST")),
-  ARGPARSE_s_s (oCompatibilityFlags, "compatibility-flags", "@"),
-  ARGPARSE_s_i (oListenBacklog, "listen-backlog", "@"),
-
-
-  ARGPARSE_header("Security", N_("Options controlling the security")),
 
   ARGPARSE_s_n (oAllowAdmin, "allow-admin", "@"),
   ARGPARSE_s_n (oDenyAdmin, "deny-admin",
                 N_("deny the use of admin card commands")),
-
+  ARGPARSE_s_s (oDisableApplication, "disable-application", "@"),
+  ARGPARSE_s_n (oEnablePinpadVarlen, "enable-pinpad-varlen",
+                N_("use variable length input for pinpad")),
+  ARGPARSE_s_s (oHomedir,    "homedir",      "@"),
+  ARGPARSE_s_i (oListenBacklog, "listen-backlog", "@"),
 
   ARGPARSE_end ()
 };
@@ -198,18 +173,8 @@ static struct debug_flags_s debug_flags [] =
     { DBG_MEMSTAT_VALUE, "memstat" },
     { DBG_HASHING_VALUE, "hashing" },
     { DBG_IPC_VALUE    , "ipc"     },
-    { DBG_CARD_VALUE   , "card"    },
     { DBG_CARD_IO_VALUE, "cardio"  },
     { DBG_READER_VALUE , "reader"  },
-    { DBG_APP_VALUE    , "app"     },
-    { 0, NULL }
-  };
-
-
-/* The list of compatibility flags.  */
-static struct compatibility_flags_s compatibility_flags [] =
-  {
-    { COMPAT_CCID_NO_AUTO_DETACH, "ccid-no-auto-detach" },
     { 0, NULL }
   };
 
@@ -249,6 +214,9 @@ static int shutdown_pending;
 /* It is possible that we are currently running under setuid permissions */
 static int maybe_setuid = 1;
 
+/* Flag telling whether we are running as a pipe server.  */
+static int pipe_server;
+
 /* Name of the communication socket */
 static char *socket_name;
 /* Name of the redirected socket or NULL.  */
@@ -279,7 +247,7 @@ static gnupg_fd_t create_server_socket (const char *name,
                                         assuan_sock_nonce_t *nonce);
 
 static void *start_connection_thread (void *arg);
-static void handle_connections (gnupg_fd_t listen_fd);
+static void handle_connections (int listen_fd);
 
 /* Pth wrapper function definitions. */
 ASSUAN_SYSTEM_NPTH_IMPL;
@@ -313,11 +281,9 @@ my_strusage (int level)
 
   switch (level)
     {
-    case  9: p = "GPL-3.0-or-later"; break;
     case 11: p = "@SCDAEMON@ (@GNUPG@)";
       break;
     case 13: p = VERSION; break;
-    case 14: p = GNUPG_DEF_COPYRIGHT_LINE; break;
     case 17: p = PRINTABLE_OS_NAME; break;
     case 19: p = _("Please report bugs to <@EMAIL@>.\n"); break;
 
@@ -427,51 +393,38 @@ cleanup (void)
     }
 }
 
-static void
-setup_signal_mask (void)
-{
-#ifndef HAVE_W32_SYSTEM
-  npth_sigev_init ();
-  npth_sigev_add (SIGHUP);
-  npth_sigev_add (SIGUSR1);
-  npth_sigev_add (SIGUSR2);
-  npth_sigev_add (SIGINT);
-  npth_sigev_add (SIGCONT);
-  npth_sigev_add (SIGTERM);
-  npth_sigev_fini ();
-  main_thread_pid = getpid ();
-#endif
-}
+
 
 int
 main (int argc, char **argv )
 {
-  gpgrt_argparse_t pargs;
+  ARGPARSE_ARGS pargs;
   int orig_argc;
   char **orig_argv;
-  char *last_configname = NULL;
-  const char *configname = NULL;
+  FILE *configfp = NULL;
+  char *configname = NULL;
   const char *shell;
-  int debug_argparser = 0;
+  unsigned int configlineno;
+  int parse_debug = 0;
   const char *debug_level = NULL;
+  int default_config =1;
   int greeting = 0;
   int nogreeting = 0;
   int multi_server = 0;
-  int pipe_server = 0;
   int is_daemon = 0;
   int nodetach = 0;
   int csh_style = 0;
   char *logfile = NULL;
   int debug_wait = 0;
   int gpgconf_list = 0;
-  char *config_filename = NULL;
+  const char *config_filename = NULL;
   int allow_coredump = 0;
   struct assuan_malloc_hooks malloc_hooks;
   int res;
-  const char *application_priority = NULL;
+  npth_t pipecon_handler;
 
   early_system_init ();
-  gpgrt_set_strusage (my_strusage);
+  set_strusage (my_strusage);
   gcry_control (GCRYCTL_SUSPEND_SECMEM_WARN);
   /* Please note that we may running SUID(ROOT), so be very CAREFUL
      when adding any stuff between here and the call to INIT_SECMEM()
@@ -511,22 +464,22 @@ main (int argc, char **argv )
   orig_argv = argv;
   pargs.argc = &argc;
   pargs.argv = &argv;
-  pargs.flags= (ARGPARSE_FLAG_KEEP | ARGPARSE_FLAG_NOVERSION);
-  while (gpgrt_argparse (NULL, &pargs, opts))
+  pargs.flags= 1|(1<<6);  /* do not remove the args, ignore version */
+  while (arg_parse( &pargs, opts))
     {
-      switch (pargs.r_opt)
-        {
-        case oDebug:
-        case oDebugAll:
-          debug_argparser++;
-          break;
-        case oHomedir:
-          gnupg_set_homedir (pargs.r.ret_str);
-          break;
+      if (pargs.r_opt == oDebug || pargs.r_opt == oDebugAll)
+        parse_debug++;
+      else if (pargs.r_opt == oOptions)
+        { /* yes there is one, so we do not try the default one, but
+             read the option file when it is encountered at the
+             commandline */
+          default_config = 0;
         }
+        else if (pargs.r_opt == oNoOptions)
+          default_config = 0; /* --no-options */
+        else if (pargs.r_opt == oHomedir)
+          gnupg_set_homedir (pargs.r.ret_str);
     }
-  /* Reset the flags.  */
-  pargs.flags &= ~(ARGPARSE_FLAG_KEEP | ARGPARSE_FLAG_NOVERSION);
 
   /* initialize the secure memory. */
   gcry_control (GCRYCTL_INIT_SECMEM, 16384, 0);
@@ -536,38 +489,48 @@ main (int argc, char **argv )
      Now we are working under our real uid
   */
 
-  /* The configuraton directories for use by gpgrt_argparser.  */
-  gpgrt_set_confdir (GPGRT_CONFDIR_SYS, gnupg_sysconfdir ());
-  gpgrt_set_confdir (GPGRT_CONFDIR_USER, gnupg_homedir ());
 
-  /* We are re-using the struct, thus the reset flag.  We OR the
-   * flags so that the internal intialized flag won't be cleared. */
+  if (default_config)
+    configname = make_filename (gnupg_homedir (), SCDAEMON_NAME EXTSEP_S "conf",
+                                NULL );
+
+
   argc = orig_argc;
   argv = orig_argv;
   pargs.argc = &argc;
   pargs.argv = &argv;
-  pargs.flags |=  (ARGPARSE_FLAG_RESET
-                   | ARGPARSE_FLAG_KEEP
-                   | ARGPARSE_FLAG_SYS
-                   | ARGPARSE_FLAG_USER);
-  while (gpgrt_argparser (&pargs, opts, SCDAEMON_NAME EXTSEP_S "conf"))
+  pargs.flags=  1;  /* do not remove the args */
+ next_pass:
+  if (configname)
+    {
+      configlineno = 0;
+      configfp = fopen (configname, "r");
+      if (!configfp)
+        {
+          if (default_config)
+            {
+              if( parse_debug )
+                log_info (_("Note: no default option file '%s'\n"),
+                          configname );
+            }
+          else
+            {
+              log_error (_("option file '%s': %s\n"),
+                         configname, strerror(errno) );
+              exit(2);
+            }
+          xfree (configname);
+          configname = NULL;
+        }
+      if (parse_debug && configname )
+        log_info (_("reading options from '%s'\n"), configname );
+      default_config = 0;
+    }
+
+  while (optfile_parse( configfp, configname, &configlineno, &pargs, opts) )
     {
       switch (pargs.r_opt)
         {
-        case ARGPARSE_CONFFILE:
-          if (debug_argparser)
-            log_info (_("reading options from '%s'\n"),
-                      pargs.r_type? pargs.r.ret_str: "[cmdline]");
-          if (pargs.r_type)
-            {
-              xfree (last_configname);
-              last_configname = xstrdup (pargs.r.ret_str);
-              configname = last_configname;
-            }
-          else
-            configname = NULL;
-          break;
-
         case aGPGConfList: gpgconf_list = 1; break;
         case aGPGConfTest: gpgconf_list = 2; break;
         case oQuiet: opt.quiet = 1; break;
@@ -596,15 +559,22 @@ main (int argc, char **argv )
         case oDebugLogTid:
           log_set_pid_suffix_cb (tid_log_callback);
           break;
-        case oDebugAllowPINLogging:
-          opt.debug_allow_pin_logging = 1;
-          break;
         case oDebugAssuanLogCats:
           set_libassuan_log_cats (pargs.r.ret_ulong);
           break;
 
+        case oOptions:
+          /* config files may not be nested (silently ignore them) */
+          if (!configfp)
+            {
+                xfree(configname);
+                configname = xstrdup(pargs.r.ret_str);
+                goto next_pass;
+            }
+          break;
         case oNoGreeting: nogreeting = 1; break;
         case oNoVerbose: opt.verbose = 0; break;
+        case oNoOptions: break; /* no-options */
         case oHomedir: gnupg_set_homedir (pargs.r.ret_str); break;
         case oNoDetach: nodetach = 1; break;
         case oLogFile: logfile = pargs.r.ret_str; break;
@@ -617,7 +587,6 @@ main (int argc, char **argv )
         case oReaderPort: opt.reader_port = pargs.r.ret_str; break;
         case octapiDriver: opt.ctapi_driver = pargs.r.ret_str; break;
         case opcscDriver: opt.pcsc_driver = pargs.r.ret_str; break;
-        case opcscShared: opt.pcsc_shared = 1; break;
         case oDisableCCID: opt.disable_ccid = 1; break;
         case oDisableOpenSC: break;
 
@@ -633,67 +602,38 @@ main (int argc, char **argv )
           add_to_strlist (&opt.disabled_applications, pargs.r.ret_str);
           break;
 
-        case oApplicationPriority:
-          application_priority = pargs.r.ret_str;
-          break;
-
         case oEnablePinpadVarlen: opt.enable_pinpad_varlen = 1; break;
 
-        case oCompatibilityFlags:
-          if (parse_compatibility_flags (pargs.r.ret_str, &opt.compat_flags,
-                                         compatibility_flags))
-            {
-              pargs.r_opt = ARGPARSE_INVALID_ARG;
-              pargs.err = ARGPARSE_PRINT_WARNING;
-            }
-          break;
-
         case oListenBacklog:
           listen_backlog = pargs.r.ret_int;
           break;
 
         default:
-          if (configname)
-            pargs.err = ARGPARSE_PRINT_WARNING;
-          else
-            pargs.err = ARGPARSE_PRINT_ERROR;
+          pargs.err = configfp? ARGPARSE_PRINT_WARNING:ARGPARSE_PRINT_ERROR;
           break;
         }
     }
-
-  gpgrt_argparse (NULL, &pargs, NULL);  /* Release internal state.  */
-
-  if (!last_configname)
-    config_filename = gpgrt_fnameconcat (gnupg_homedir (),
-                                         SCDAEMON_NAME EXTSEP_S "conf",
-                                         NULL);
-  else
+  if (configfp)
     {
-      config_filename = last_configname;
-      last_configname = NULL;
+      fclose( configfp );
+      configfp = NULL;
+      /* Keep a copy of the config name for use by --gpgconf-list. */
+      config_filename = configname;
+      configname = NULL;
+      goto next_pass;
     }
-
+  xfree (configname);
+  configname = NULL;
   if (log_get_errorcount(0))
     exit(2);
-
-  /* Process common component options.  */
-  if (parse_comopt (GNUPG_MODULE_NAME_SCDAEMON, debug_argparser))
-    exit(2);
-
-  if (!logfile)
-    {
-      logfile = comopt.logfile;
-      comopt.logfile = NULL;
-    }
-
   if (nogreeting )
     greeting = 0;
 
   if (greeting)
     {
       es_fprintf (es_stderr, "%s %s; %s\n",
-                  gpgrt_strusage (11),gpgrt_strusage (13),gpgrt_strusage (14));
-      es_fprintf (es_stderr, "%s\n", gpgrt_strusage (15));
+                  strusage(11), strusage(13), strusage(14) );
+      es_fprintf (es_stderr, "%s\n", strusage(15) );
     }
 #ifdef IS_DEVELOPMENT_VERSION
   log_info ("NOTE: this is a development version!\n");
@@ -730,10 +670,42 @@ main (int argc, char **argv )
   if (gpgconf_list)
     {
       /* List options and default values in the GPG Conf format.  */
-      es_printf ("debug-level:%lu:\"none:\n", GC_OPT_FLAG_DEFAULT);
+      char *filename = NULL;
+      char *filename_esc;
+
+      if (config_filename)
+        filename = xstrdup (config_filename);
+      else
+        filename = make_filename (gnupg_homedir (),
+                                  SCDAEMON_NAME EXTSEP_S "conf", NULL);
+      filename_esc = percent_escape (filename, NULL);
+
+      es_printf ("%s-%s.conf:%lu:\"%s\n",
+                 GPGCONF_NAME, SCDAEMON_NAME,
+                 GC_OPT_FLAG_DEFAULT, filename_esc);
+      xfree (filename_esc);
+      xfree (filename);
+
+      es_printf ("verbose:%lu:\n"
+                 "quiet:%lu:\n"
+                 "debug-level:%lu:\"none:\n"
+                 "log-file:%lu:\n",
+                 GC_OPT_FLAG_NONE,
+                 GC_OPT_FLAG_NONE,
+                 GC_OPT_FLAG_DEFAULT,
+                 GC_OPT_FLAG_NONE );
+
+      es_printf ("reader-port:%lu:\n", GC_OPT_FLAG_NONE );
+      es_printf ("ctapi-driver:%lu:\n", GC_OPT_FLAG_NONE );
       es_printf ("pcsc-driver:%lu:\"%s:\n",
-                 GC_OPT_FLAG_DEFAULT, DEFAULT_PCSC_DRIVER );
+              GC_OPT_FLAG_DEFAULT, DEFAULT_PCSC_DRIVER );
+#ifdef HAVE_LIBUSB
+      es_printf ("disable-ccid:%lu:\n", GC_OPT_FLAG_NONE );
+#endif
+      es_printf ("deny-admin:%lu:\n", GC_OPT_FLAG_NONE );
+      es_printf ("disable-pinpad:%lu:\n", GC_OPT_FLAG_NONE );
       es_printf ("card-timeout:%lu:%d:\n", GC_OPT_FLAG_DEFAULT, 0);
+      es_printf ("enable-pinpad-varlen:%lu:\n", GC_OPT_FLAG_NONE );
 
       scd_exit (0);
     }
@@ -742,9 +714,7 @@ main (int argc, char **argv )
   if (logfile)
     {
       log_set_file (logfile);
-      log_set_prefix (NULL, (GPGRT_LOG_WITH_PREFIX
-                             | GPGRT_LOG_WITH_TIME
-                             | GPGRT_LOG_WITH_PID));
+      log_set_prefix (NULL, GPGRT_LOG_WITH_PREFIX | GPGRT_LOG_WITH_TIME | GPGRT_LOG_WITH_PID);
     }
 
   if (debug_wait && pipe_server)
@@ -755,16 +725,12 @@ main (int argc, char **argv )
       log_debug ("... okay\n");
     }
 
-  if (application_priority)
-    app_update_priority_list (application_priority);
-
   if (pipe_server)
     {
       /* This is the simple pipe based server */
       ctrl_t ctrl;
       npth_attr_t tattr;
-      gnupg_fd_t fd = GNUPG_INVALID_FD;
-      npth_t pipecon_handler;
+      int fd = -1;
 
 #ifndef HAVE_W32_SYSTEM
       {
@@ -778,7 +744,6 @@ main (int argc, char **argv )
 #endif
 
       npth_init ();
-      setup_signal_mask ();
       gpgrt_set_syscall_clamp (npth_unprotect, npth_protect);
 
       /* If --debug-allow-core-dump has been given we also need to
@@ -799,8 +764,8 @@ main (int argc, char **argv )
       if (multi_server)
         {
           socket_name = create_socket_name (SCDAEMON_SOCK_NAME);
-          fd = create_server_socket (socket_name,
-                                     &redir_socket_name, &socket_nonce);
+          fd = FD2INT(create_server_socket (socket_name,
+                                            &redir_socket_name, &socket_nonce));
         }
 
       res = npth_attr_init (&tattr);
@@ -834,8 +799,8 @@ main (int argc, char **argv )
       /* We run handle_connection to wait for the shutdown signal and
          to run the ticker stuff.  */
       handle_connections (fd);
-      if (fd != GNUPG_INVALID_FD)
-        assuan_sock_close (fd);
+      if (fd != -1)
+        close (fd);
     }
   else if (!is_daemon)
     {
@@ -844,7 +809,7 @@ main (int argc, char **argv )
     }
   else
     { /* Regular server mode */
-      gnupg_fd_t fd;
+      int fd;
 #ifndef HAVE_W32_SYSTEM
       pid_t pid;
       int i;
@@ -852,8 +817,8 @@ main (int argc, char **argv )
 
       /* Create the socket.  */
       socket_name = create_socket_name (SCDAEMON_SOCK_NAME);
-      fd = create_server_socket (socket_name,
-                                 &redir_socket_name, &socket_nonce);
+      fd = FD2INT (create_server_socket (socket_name,
+                                         &redir_socket_name, &socket_nonce));
 
 
       fflush (NULL);
@@ -919,7 +884,6 @@ main (int argc, char **argv )
       /* This is the child. */
 
       npth_init ();
-      setup_signal_mask ();
       gpgrt_set_syscall_clamp (npth_unprotect, npth_protect);
 
       /* Detach from tty and put process into a new session. */
@@ -969,10 +933,9 @@ main (int argc, char **argv )
 
       handle_connections (fd);
 
-      assuan_sock_close (fd);
+      close (fd);
     }
 
-  xfree (config_filename);
   return 0;
 }
 
@@ -1020,7 +983,7 @@ scd_deinit_default_ctrl (ctrl_t ctrl)
 /* Return the name of the socket to be used to connect to this
    process.  If no socket is available, return NULL. */
 const char *
-scd_get_socket_name (void)
+scd_get_socket_name ()
 {
   if (socket_name && *socket_name)
     return socket_name;
@@ -1028,36 +991,7 @@ scd_get_socket_name (void)
 }
 
 
-#ifdef HAVE_W32_SYSTEM
-/* Creat an event into E_P and sets EVENTS to watch by npth_eselect.  */
-void
-scd_init_event (HANDLE *e_p, HANDLE events[2])
-{
-  HANDLE h, h2;
-  SECURITY_ATTRIBUTES sa = { sizeof (SECURITY_ATTRIBUTES), NULL, TRUE};
-
-  events[0] = *e_p = INVALID_HANDLE_VALUE;
-  events[1] = INVALID_HANDLE_VALUE;
-  /* Create event for manual reset, initially non-signaled.  Make it
-   * waitable and inheritable.  */
-  h = CreateEvent (&sa, TRUE, FALSE, NULL);
-  if (!h)
-    log_error ("can't create scd event: %s\n", w32_strerror (-1) );
-  else if (!DuplicateHandle (GetCurrentProcess(), h,
-                             GetCurrentProcess(), &h2,
-                             EVENT_MODIFY_STATE|SYNCHRONIZE, TRUE, 0))
-    {
-      log_error ("setting synchronize for scd_kick_the_loop failed: %s\n",
-                 w32_strerror (-1) );
-      CloseHandle (h);
-    }
-  else
-    {
-      CloseHandle (h);
-      events[0] = *e_p = h2;
-    }
-}
-#else
+#ifndef HAVE_W32_SYSTEM
 static void
 handle_signal (int signo)
 {
@@ -1096,7 +1030,7 @@ handle_signal (int signo)
       if (shutdown_pending > 2)
         {
           log_info ("shutdown forced\n");
-          log_info ("%s %s stopped\n", gpgrt_strusage(11), gpgrt_strusage(13));
+          log_info ("%s %s stopped\n", strusage(11), strusage(13) );
           cleanup ();
           scd_exit (0);
         }
@@ -1104,7 +1038,7 @@ handle_signal (int signo)
 
     case SIGINT:
       log_info ("SIGINT received - immediate shutdown\n");
-      log_info( "%s %s stopped\n", gpgrt_strusage(11), gpgrt_strusage(13));
+      log_info( "%s %s stopped\n", strusage(11), strusage(13));
       cleanup ();
       scd_exit (0);
       break;
@@ -1119,7 +1053,7 @@ handle_signal (int signo)
 /* Create a name for the socket.  We check for valid characters as
    well as against a maximum allowed length for a unix domain socket
    is done.  The function terminates the process in case of an error.
-   Returns: Pointer to an allocated string with the absolute name of
+   Retunrs: Pointer to an allcoated string with the absolute name of
    the socket used.  */
 static char *
 create_socket_name (char *standard_name)
@@ -1247,13 +1181,12 @@ start_connection_thread (void *arg)
     log_info (_("handler for fd %d started\n"),
               FD2INT(ctrl->thread_startup.fd));
 
-  scd_command_handler (ctrl, ctrl->thread_startup.fd);
-
-  /* If this thread is the pipe connection thread, flag that a
-     shutdown is required.  With the next ticker event and given that
+  /* If this is a pipe server, we request a shutdown if the command
+     handler asked for it.  With the next ticker event and given that
      no other connections are running the shutdown will then
      happen.  */
-  if (ctrl->thread_startup.fd == GNUPG_INVALID_FD)
+  if (scd_command_handler (ctrl, FD2INT(ctrl->thread_startup.fd))
+      && pipe_server)
     shutdown_pending = 1;
 
   if (opt.verbose)
@@ -1273,18 +1206,20 @@ start_connection_thread (void *arg)
 void
 scd_kick_the_loop (void)
 {
+  int ret;
+
   /* Kick the select loop.  */
 #ifdef HAVE_W32_SYSTEM
-  int ret = SetEvent (the_event);
+  ret = SetEvent (the_event);
   if (ret == 0)
     log_error ("SetEvent for scd_kick_the_loop failed: %s\n",
                w32_strerror (-1));
 #elif defined(HAVE_PSELECT_NO_EINTR)
   write (notify_fd, "", 1);
 #else
-  int ret = kill (main_thread_pid, SIGCONT);
+  ret = kill (main_thread_pid, SIGCONT);
   if (ret < 0)
-    log_error ("sending signal for scd_kick_the_loop failed: %s\n",
+    log_error ("SetEvent for scd_kick_the_loop failed: %s\n",
                gpg_strerror (gpg_error_from_syserror ()));
 #endif
 }
@@ -1294,7 +1229,7 @@ scd_kick_the_loop (void)
    in which case this code will only do regular timeouts and handle
    signals. */
 static void
-handle_connections (gnupg_fd_t listen_fd)
+handle_connections (int listen_fd)
 {
   npth_attr_t tattr;
   struct sockaddr_un paddr;
@@ -1302,6 +1237,7 @@ handle_connections (gnupg_fd_t listen_fd)
   fd_set fdset, read_fdset;
   int nfd;
   int ret;
+  int fd;
   struct timespec timeout;
   struct timespec *t;
   int saved_errno;
@@ -1333,15 +1269,47 @@ handle_connections (gnupg_fd_t listen_fd)
   npth_attr_setdetachstate (&tattr, NPTH_CREATE_DETACHED);
 
 #ifdef HAVE_W32_SYSTEM
-  scd_init_event (&the_event, events);
+  {
+    HANDLE h, h2;
+    SECURITY_ATTRIBUTES sa = { sizeof (SECURITY_ATTRIBUTES), NULL, TRUE};
+
+    events[0] = the_event = INVALID_HANDLE_VALUE;
+    events[1] = INVALID_HANDLE_VALUE;
+    h = CreateEvent (&sa, TRUE, FALSE, NULL);
+    if (!h)
+      log_error ("can't create scd event: %s\n", w32_strerror (-1) );
+    else if (!DuplicateHandle (GetCurrentProcess(), h,
+                               GetCurrentProcess(), &h2,
+                               EVENT_MODIFY_STATE|SYNCHRONIZE, TRUE, 0))
+      {
+        log_error ("setting synchronize for scd_kick_the_loop failed: %s\n",
+                   w32_strerror (-1) );
+        CloseHandle (h);
+      }
+    else
+      {
+        CloseHandle (h);
+        events[0] = the_event = h2;
+      }
+  }
+#else
+  npth_sigev_init ();
+  npth_sigev_add (SIGHUP);
+  npth_sigev_add (SIGUSR1);
+  npth_sigev_add (SIGUSR2);
+  npth_sigev_add (SIGINT);
+  npth_sigev_add (SIGCONT);
+  npth_sigev_add (SIGTERM);
+  npth_sigev_fini ();
+  main_thread_pid = getpid ();
 #endif
 
   FD_ZERO (&fdset);
   nfd = 0;
-  if (listen_fd != GNUPG_INVALID_FD)
+  if (listen_fd != -1)
     {
-      FD_SET (FD2INT (listen_fd), &fdset);
-      nfd = FD2INT (listen_fd);
+      FD_SET (listen_fd, &fdset);
+      nfd = listen_fd;
     }
 
   for (;;)
@@ -1359,7 +1327,7 @@ handle_connections (gnupg_fd_t listen_fd)
              file descriptors to wait for, so that the select will be
              used to just wait on a signal or timeout event. */
           FD_ZERO (&fdset);
-          listen_fd = GNUPG_INVALID_FD;
+          listen_fd = -1;
         }
 
       periodical_check = scd_update_reader_status_file ();
@@ -1380,8 +1348,6 @@ handle_connections (gnupg_fd_t listen_fd)
       FD_SET (pipe_fd[0], &read_fdset);
       if (max_fd < pipe_fd[0])
         max_fd = pipe_fd[0];
-#else
-      (void)max_fd;
 #endif
 
 #ifndef HAVE_W32_SYSTEM
@@ -1395,18 +1361,15 @@ handle_connections (gnupg_fd_t listen_fd)
       ret = npth_eselect (nfd+1, &read_fdset, NULL, NULL, t,
                           events, &events_set);
       saved_errno = errno;
-      if ((events_set & 1))
-        {
-          if (opt.debug)
-            log_debug ("the_event seen and reset by npth_eselect\n");
-        }
+      if (events_set & 1)
+        continue;
 #endif
 
       if (ret == -1 && saved_errno != EINTR)
         {
           log_error (_("npth_pselect failed: %s - waiting 1s\n"),
                      strerror (saved_errno));
-          gnupg_sleep (1);
+          npth_sleep (1);
           continue;
         }
 
@@ -1423,16 +1386,13 @@ handle_connections (gnupg_fd_t listen_fd)
         }
 #endif
 
-      if (listen_fd != GNUPG_INVALID_FD
-          && FD_ISSET (FD2INT (listen_fd), &read_fdset))
+      if (listen_fd != -1 && FD_ISSET (listen_fd, &read_fdset))
         {
           ctrl_t ctrl;
-          gnupg_fd_t fd;
 
           plen = sizeof paddr;
-          fd = INT2FD (npth_accept (FD2INT (listen_fd),
-                                    (struct sockaddr *)&paddr, &plen));
-          if (fd == GNUPG_INVALID_FD)
+          fd = npth_accept (listen_fd, (struct sockaddr *)&paddr, &plen);
+          if (fd == -1)
             {
               log_error ("accept failed: %s\n", strerror (errno));
             }
@@ -1440,23 +1400,22 @@ handle_connections (gnupg_fd_t listen_fd)
             {
               log_error ("error allocating connection control data: %s\n",
                          strerror (errno) );
-              assuan_sock_close (fd);
+              close (fd);
             }
           else
             {
               char threadname[50];
               npth_t thread;
 
-              snprintf (threadname, sizeof threadname, "conn fd=%d",
-                        FD2INT (fd));
-              ctrl->thread_startup.fd = fd;
+              snprintf (threadname, sizeof threadname, "conn fd=%d", fd);
+              ctrl->thread_startup.fd = INT2FD (fd);
               ret = npth_create (&thread, &tattr, start_connection_thread, ctrl);
               if (ret)
                 {
                   log_error ("error spawning connection handler: %s\n",
                              strerror (ret));
                   xfree (ctrl);
-                  assuan_sock_close (fd);
+                  close (fd);
                 }
               else
                 npth_setname_np (thread, threadname);
@@ -1473,7 +1432,7 @@ handle_connections (gnupg_fd_t listen_fd)
   close (pipe_fd[1]);
 #endif
   cleanup ();
-  log_info (_("%s %s stopped\n"), gpgrt_strusage(11), gpgrt_strusage(13));
+  log_info (_("%s %s stopped\n"), strusage(11), strusage(13));
   npth_attr_destroy (&tattr);
 }
 
