@@ -16,7 +16,6 @@
  *
  * You should have received a copy of the GNU Lesser General Public License
  * along with this program; if not, see <https://www.gnu.org/licenses/>.
- * SPDX-License-Identifier: LGPL-2.1-or-later
  */
 
 /* The Web Key Service I-D defines an update protocol to store a
@@ -25,7 +24,6 @@
  */
 
 #include <config.h>
-
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -34,7 +32,6 @@
 #include <sys/stat.h>
 #include <dirent.h>
 
-#define INCLUDED_BY_MAIN_MODULE 1
 #include "../common/util.h"
 #include "../common/init.h"
 #include "../common/sysutils.h"
@@ -61,7 +58,6 @@ enum cmd_and_opt_values
     oQuiet      = 'q',
     oVerbose	= 'v',
     oOutput     = 'o',
-    oDirectory  = 'C',
 
     oDebug      = 500,
 
@@ -85,7 +81,7 @@ enum cmd_and_opt_values
 
 
 /* The list of commands and options. */
-static gpgrt_opt_t opts[] = {
+static ARGPARSE_OPTS opts[] = {
   ARGPARSE_group (300, ("@Commands:\n ")),
 
   ARGPARSE_c (aReceive,   "receive",
@@ -112,7 +108,6 @@ static gpgrt_opt_t opts[] = {
   ARGPARSE_s_s (oGpgProgram, "gpg", "@"),
   ARGPARSE_s_n (oSend, "send", "send the mail using sendmail"),
   ARGPARSE_s_s (oOutput, "output", "|FILE|write the mail to FILE"),
-  ARGPARSE_s_s (oDirectory, "directory", "|DIR|use DIR as top directory"),
   ARGPARSE_s_s (oFrom, "from", "|ADDR|use ADDR as the default sender"),
   ARGPARSE_s_s (oHeader, "header" ,
                 "|NAME=VALUE|add \"NAME: VALUE\" as header to all mails"),
@@ -160,6 +155,8 @@ static gpg_error_t command_receive_cb (void *opaque,
                                        const char *mediatype, estream_t fp,
                                        unsigned int flags);
 static gpg_error_t command_list_domains (void);
+static gpg_error_t command_install_key (const char *fname, const char *userid);
+static gpg_error_t command_remove_key (const char *mailaddr);
 static gpg_error_t command_revoke_key (const char *mailaddr);
 static gpg_error_t command_check_key (const char *mailaddr);
 static gpg_error_t command_cron (void);
@@ -174,11 +171,9 @@ my_strusage( int level )
 
   switch (level)
     {
-    case  9: p = "LGPL-2.1-or-later"; break;
     case 11: p = "gpg-wks-server"; break;
     case 12: p = "@GNUPG@"; break;
     case 13: p = VERSION; break;
-    case 14: p = GNUPG_DEF_COPYRIGHT_LINE; break;
     case 17: p = PRINTABLE_OS_NAME; break;
     case 19: p = ("Please report bugs to <@EMAIL@>.\n"); break;
 
@@ -200,7 +195,7 @@ my_strusage( int level )
 static void
 wrong_args (const char *text)
 {
-  es_fprintf (es_stderr, "usage: %s [options] %s\n", gpgrt_strusage (11), text);
+  es_fprintf (es_stderr, "usage: %s [options] %s\n", strusage (11), text);
   exit (2);
 }
 
@@ -208,12 +203,12 @@ wrong_args (const char *text)
 
 /* Command line parsing.  */
 static enum cmd_and_opt_values
-parse_arguments (gpgrt_argparse_t *pargs, gpgrt_opt_t *popts)
+parse_arguments (ARGPARSE_ARGS *pargs, ARGPARSE_OPTS *popts)
 {
   enum cmd_and_opt_values cmd = 0;
   int no_more_options = 0;
 
-  while (!no_more_options && gpgrt_argparse (NULL, pargs, popts))
+  while (!no_more_options && optfile_parse (NULL, NULL, NULL, pargs, popts))
     {
       switch (pargs->r_opt)
         {
@@ -230,9 +225,6 @@ parse_arguments (gpgrt_argparse_t *pargs, gpgrt_opt_t *popts)
         case oGpgProgram:
           opt.gpg_program = pargs->r.ret_str;
           break;
-        case oDirectory:
-          opt.directory = pargs->r.ret_str;
-          break;
         case oFrom:
           opt.default_from = pargs->r.ret_str;
           break;
@@ -262,7 +254,7 @@ parse_arguments (gpgrt_argparse_t *pargs, gpgrt_opt_t *popts)
           cmd = pargs->r_opt;
           break;
 
-        default: pargs->err = ARGPARSE_PRINT_ERROR; break;
+        default: pargs->err = 2; break;
 	}
     }
 
@@ -276,11 +268,11 @@ int
 main (int argc, char **argv)
 {
   gpg_error_t err, firsterr;
-  gpgrt_argparse_t pargs;
+  ARGPARSE_ARGS pargs;
   enum cmd_and_opt_values cmd;
 
   gnupg_reopen_std ("gpg-wks-server");
-  gpgrt_set_strusage (my_strusage);
+  set_strusage (my_strusage);
   log_set_prefix ("gpg-wks-server", GPGRT_LOG_WITH_PREFIX);
 
   /* Make sure that our subsystems are ready.  */
@@ -291,7 +283,6 @@ main (int argc, char **argv)
   pargs.argv  = &argv;
   pargs.flags = ARGPARSE_FLAG_KEEP;
   cmd = parse_arguments (&pargs, opts);
-  gpgrt_argparse (NULL, &pargs, NULL);
 
   if (log_get_errorcount (0))
     exit (2);
@@ -337,7 +328,7 @@ main (int argc, char **argv)
   {
     struct stat sb;
 
-    if (gnupg_stat (opt.directory, &sb))
+    if (stat (opt.directory, &sb))
       {
         err = gpg_error_from_syserror ();
         log_error ("error accessing directory '%s': %s\n",
@@ -359,7 +350,6 @@ main (int argc, char **argv)
       {
         log_error ("directory '%s' has too relaxed permissions\n",
                    opt.directory);
-        log_info ("Fix by running: chmod o-rw '%s'\n", opt.directory);
         exit (2);
       }
   }
@@ -387,18 +377,15 @@ main (int argc, char **argv)
       break;
 
     case aInstallKey:
-      if (!argc)
-        err = wks_cmd_install_key (NULL, NULL);
-      else if (argc == 2)
-        err = wks_cmd_install_key (*argv, argv[1]);
-      else
-        wrong_args ("--install-key [FILE|FINGERPRINT USER-ID]");
+      if (argc != 2)
+        wrong_args ("--install-key FILE USER-ID");
+      err = command_install_key (*argv, argv[1]);
       break;
 
     case aRemoveKey:
       if (argc != 1)
         wrong_args ("--remove-key USER-ID");
-      err = wks_cmd_remove_key (*argv);
+      err = command_remove_key (*argv);
       break;
 
     case aRevokeKey:
@@ -421,7 +408,7 @@ main (int argc, char **argv)
       break;
 
     default:
-      gpgrt_usage (1);
+      usage (1);
       err = gpg_error (GPG_ERR_BUG);
       break;
     }
@@ -592,7 +579,6 @@ encrypt_stream (estream_t *r_output, estream_t input, const char *keyfile)
   ccparray_put (&ccp, "--always-trust");
   ccparray_put (&ccp, "--no-keyring");
   ccparray_put (&ccp, "--armor");
-  ccparray_put (&ccp, "-z0");  /* No compression for improved robustness.  */
   ccparray_put (&ccp, "--recipient-file");
   ccparray_put (&ccp, keyfile);
   ccparray_put (&ccp, "--encrypt");
@@ -809,65 +795,17 @@ get_policy_flags (policy_flags_t policy, const char *mbox)
 }
 
 
-/* Create the name for the pending file from NONCE and ADDRSPEC and
- * store it at R_NAME.  */
-static gpg_error_t
-make_pending_fname (const char *nonce, const char *addrspec, char **r_name)
-{
-  gpg_error_t err = 0;
-  const char *domain;
-  char *addrspechash = NULL;
-  char sha1buf[20];
-
-  *r_name = NULL;
-
-  domain = addrspec? strchr (addrspec, '@') : NULL;
-  if (!domain || !domain[1] || domain == addrspec)
-    {
-      err = gpg_error (GPG_ERR_INV_ARG);
-      goto leave;
-    }
-  domain++;
-  if (strchr (domain, '/') || strchr (domain, '\\'))
-    {
-      log_info ("invalid domain detected ('%s')\n", domain);
-      err = gpg_error (GPG_ERR_NOT_FOUND);
-      goto leave;
-    }
-  gcry_md_hash_buffer (GCRY_MD_SHA1, sha1buf, addrspec, domain - addrspec - 1);
-  addrspechash = zb32_encode (sha1buf, 8*20);
-  if (!addrspechash)
-    {
-      err = gpg_error_from_syserror ();
-      goto leave;
-    }
-
-  *r_name = strconcat (nonce, ".", addrspechash, NULL);
-  if (!*r_name)
-    {
-      err = gpg_error_from_syserror ();
-      goto leave;
-    }
-
- leave:
-  xfree (addrspechash);
-  return err;
-}
-
-
 /* We store the key under the name of the nonce we will then send to
  * the user.  On success the nonce is stored at R_NONCE and the file
- * name at R_FNAME.  ADDRSPEC is used as part of the pending file name
- * so that the nonce is associated with an address */
+ * name at R_FNAME.  */
 static gpg_error_t
-store_key_as_pending (const char *dir, estream_t key, const char *addrspec,
+store_key_as_pending (const char *dir, estream_t key,
                       char **r_nonce, char **r_fname)
 {
   gpg_error_t err;
   char *dname = NULL;
   char *fname = NULL;
   char *nonce = NULL;
-  char *pendingname = NULL;
   estream_t outfp = NULL;
   char buffer[1024];
   size_t nbytes, nwritten;
@@ -896,11 +834,7 @@ store_key_as_pending (const char *dir, estream_t key, const char *addrspec,
       goto leave;
     }
 
-  err = make_pending_fname (nonce, addrspec, &pendingname);
-  if (err)
-    goto leave;
-
-  fname = strconcat (dname, "/", pendingname, NULL);
+  fname = strconcat (dname, "/", nonce, NULL);
   if (!fname)
     {
       err = gpg_error_from_syserror ();
@@ -969,7 +903,6 @@ store_key_as_pending (const char *dir, estream_t key, const char *addrspec,
       xfree (fname);
     }
   xfree (dname);
-  xfree (pendingname);
   return err;
 }
 
@@ -1235,7 +1168,7 @@ process_new_key (server_ctx_t ctx, estream_t key)
           goto leave;
         }
 
-      if (gnupg_access (dname, W_OK))
+      if (access (dname, W_OK))
         {
           log_info ("skipping address '%s': Domain not configured\n", sl->mbox);
           continue;
@@ -1260,7 +1193,7 @@ process_new_key (server_ctx_t ctx, estream_t key)
 
           xfree (nonce);
           xfree (fname);
-          err = store_key_as_pending (dname, key, sl->mbox, &nonce, &fname);
+          err = store_key_as_pending (dname, key, &nonce, &fname);
           if (err)
             goto leave;
 
@@ -1333,10 +1266,10 @@ send_congratulation_message (const char *mbox, const char *keyfile)
               "\n"
               "Best regards\n"
               "\n"
-              "  GnuPG Key Publisher\n\n\n"
+              "  Gnu Key Publisher\n\n\n"
               "-- \n"
-              "For information on GnuPG see: %s\n",
-              mbox, "https://gnupg.org");
+              "The GnuPG Project welcomes donations: %s\n",
+              mbox, "https://gnupg.org/donate");
 
   es_rewind (body);
   err = encrypt_stream (&bodyenc, body, keyfile);
@@ -1407,6 +1340,81 @@ send_congratulation_message (const char *mbox, const char *keyfile)
 }
 
 
+/* Write the content of SRC to the new file FNAME.  */
+static gpg_error_t
+write_to_file (estream_t src, const char *fname)
+{
+  gpg_error_t err;
+  estream_t dst;
+  char buffer[4096];
+  size_t nread, written;
+
+  dst = es_fopen (fname, "wb");
+  if (!dst)
+    return gpg_error_from_syserror ();
+
+  do
+    {
+      nread = es_fread (buffer, 1, sizeof buffer, src);
+      if (!nread)
+	break;
+      written = es_fwrite (buffer, 1, nread, dst);
+      if (written != nread)
+	break;
+    }
+  while (!es_feof (src) && !es_ferror (src) && !es_ferror (dst));
+  if (!es_feof (src) || es_ferror (src) || es_ferror (dst))
+    {
+      err = gpg_error_from_syserror ();
+      es_fclose (dst);
+      gnupg_remove (fname);
+      return err;
+    }
+
+  if (es_fclose (dst))
+    {
+      err = gpg_error_from_syserror ();
+      log_error ("error closing '%s': %s\n", fname, gpg_strerror (err));
+      return err;
+    }
+
+  return 0;
+}
+
+
+/* Compute the the full file name for the key with ADDRSPEC and return
+ * it at R_FNAME.  */
+static gpg_error_t
+compute_hu_fname (char **r_fname, const char *addrspec)
+{
+  gpg_error_t err;
+  char *hash;
+  const char *domain;
+  char sha1buf[20];
+
+  *r_fname = NULL;
+
+  domain = strchr (addrspec, '@');
+  if (!domain || !domain[1] || domain == addrspec)
+    return gpg_error (GPG_ERR_INV_ARG);
+  domain++;
+
+  gcry_md_hash_buffer (GCRY_MD_SHA1, sha1buf, addrspec, domain - addrspec - 1);
+  hash = zb32_encode (sha1buf, 8*20);
+  if (!hash)
+    return gpg_error_from_syserror ();
+
+  *r_fname = make_filename_try (opt.directory, domain, "hu", hash, NULL);
+  if (!*r_fname)
+    err = gpg_error_from_syserror ();
+  else
+    err = 0;
+
+  xfree (hash);
+  return err;
+}
+
+
 /* Check that we have send a request with NONCE and publish the key.  */
 static gpg_error_t
 check_and_publish (server_ctx_t ctx, const char *address, const char *nonce)
@@ -1416,7 +1424,6 @@ check_and_publish (server_ctx_t ctx, const char *address, const char *nonce)
   char *fnewname = NULL;
   estream_t key = NULL;
   char *hash = NULL;
-  char *pendingname = NULL;
   const char *domain;
   const char *s;
   uidinfo_list_t sl;
@@ -1433,21 +1440,7 @@ check_and_publish (server_ctx_t ctx, const char *address, const char *nonce)
   domain = strchr (address, '@');
   log_assert (domain && domain[1]);
   domain++;
-  if (strchr (domain, '/') || strchr (domain, '\\')
-      || strchr (nonce, '/') || strchr (nonce, '\\'))
-    {
-      log_info ("invalid domain or nonce received ('%s', '%s')\n",
-                domain, nonce);
-      err = gpg_error (GPG_ERR_NOT_FOUND);
-      goto leave;
-    }
-
-  err = make_pending_fname (nonce, address, &pendingname);
-  if (err)
-    goto leave;
-
-  fname = make_filename_try (opt.directory, domain, "pending", pendingname,
-                             NULL);
+  fname = make_filename_try (opt.directory, domain, "pending", nonce, NULL);
   if (!fname)
     {
       err = gpg_error_from_syserror ();
@@ -1496,7 +1489,7 @@ check_and_publish (server_ctx_t ctx, const char *address, const char *nonce)
     }
 
   /* Hash user ID and create filename.  */
-  err = wks_compute_hu_fname (&fnewname, address);
+  err = compute_hu_fname (&fnewname, address);
   if (err)
     goto leave;
 
@@ -1511,7 +1504,7 @@ check_and_publish (server_ctx_t ctx, const char *address, const char *nonce)
     }
 
   /* Make sure it is world readable.  */
-  if (gnupg_chmod (fnewname, "-rw-r--r--"))
+  if (gnupg_chmod (fnewname, "-rwxr--r--"))
     log_error ("can't set permissions of '%s': %s\n",
                fnewname, gpg_strerror (gpg_err_code_from_syserror()));
 
@@ -1527,7 +1520,7 @@ check_and_publish (server_ctx_t ctx, const char *address, const char *nonce)
       err = gpg_error_from_syserror ();
       goto leave;
     }
-  if (!gnupg_access (fnewname, W_OK))
+  if (!access (fnewname, W_OK))
     {
       /* Yes, we have a dane directory.  */
       s = strchr (address, '@');
@@ -1559,7 +1552,6 @@ check_and_publish (server_ctx_t ctx, const char *address, const char *nonce)
   xfree (fnewname);
   xfree (fname);
   xfree (nonce2);
-  xfree (pendingname);
   return err;
 }
 
@@ -1675,29 +1667,29 @@ command_receive_cb (void *opaque, const char *mediatype,
 
 
 
-/* Return a list of all configured domains.  Each list element is the
+/* Return a list of all configured domains.  ECh list element is the
  * top directory for the domain.  To figure out the actual domain
  * name strrchr(name, '/') can be used.  */
 static gpg_error_t
 get_domain_list (strlist_t *r_list)
 {
   gpg_error_t err;
-  gnupg_dir_t dir = NULL;
+  DIR *dir = NULL;
   char *fname = NULL;
-  gnupg_dirent_t dentry;
+  struct dirent *dentry;
   struct stat sb;
   strlist_t list = NULL;
 
   *r_list = NULL;
 
-  dir = gnupg_opendir (opt.directory);
+  dir = opendir (opt.directory);
   if (!dir)
     {
       err = gpg_error_from_syserror ();
       goto leave;
     }
 
-  while ((dentry = gnupg_readdir (dir)))
+  while ((dentry = readdir (dir)))
     {
       if (*dentry->d_name == '.')
         continue;
@@ -1714,7 +1706,7 @@ get_domain_list (strlist_t *r_list)
           goto leave;
         }
 
-      if (gnupg_stat (fname, &sb))
+      if (stat (fname, &sb))
         {
           err = gpg_error_from_syserror ();
           log_error ("error accessing '%s': %s\n", fname, gpg_strerror (err));
@@ -1737,7 +1729,8 @@ get_domain_list (strlist_t *r_list)
 
  leave:
   free_strlist (list);
-  gnupg_closedir (dir);
+  if (dir)
+    closedir (dir);
   xfree (fname);
   return err;
 }
@@ -1750,8 +1743,8 @@ expire_one_domain (const char *top_dirname, const char *domain)
   gpg_error_t err;
   char *dirname;
   char *fname = NULL;
-  gnupg_dir_t dir = NULL;
-  gnupg_dirent_t dentry;
+  DIR *dir = NULL;
+  struct dirent *dentry;
   struct stat sb;
   time_t now = gnupg_get_time ();
 
@@ -1764,7 +1757,7 @@ expire_one_domain (const char *top_dirname, const char *domain)
       goto leave;
     }
 
-  dir = gnupg_opendir (dirname);
+  dir = opendir (dirname);
   if (!dir)
     {
       err = gpg_error_from_syserror ();
@@ -1773,7 +1766,7 @@ expire_one_domain (const char *top_dirname, const char *domain)
       goto leave;
     }
 
-  while ((dentry = gnupg_readdir (dir)))
+  while ((dentry = readdir (dir)))
     {
       if (*dentry->d_name == '.')
         continue;
@@ -1786,13 +1779,12 @@ expire_one_domain (const char *top_dirname, const char *domain)
                      __func__, gpg_strerror (err));
           goto leave;
         }
-      /* The old files are 32 bytes, those created since 2.3.8 are 65 bytes. */
-      if (strlen (dentry->d_name) != 32 && strlen (dentry->d_name) != 65)
+      if (strlen (dentry->d_name) != 32)
         {
           log_info ("garbage file '%s' ignored\n", fname);
           continue;
         }
-      if (gnupg_stat (fname, &sb))
+      if (stat (fname, &sb))
         {
           err = gpg_error_from_syserror ();
           log_error ("error accessing '%s': %s\n", fname, gpg_strerror (err));
@@ -1823,7 +1815,8 @@ expire_one_domain (const char *top_dirname, const char *domain)
   err = 0;
 
  leave:
-  gnupg_closedir (dir);
+  if (dir)
+    closedir (dir);
   xfree (dirname);
   xfree (fname);
   return err;
@@ -1863,7 +1856,7 @@ command_list_domains (void)
     { "pending", "-rwx" },
     { "hu",      "-rwxr-xr-x" }
   };
-  gpg_err_code_t ec;
+
   gpg_error_t err;
   strlist_t domaindirs;
   strlist_t sl;
@@ -1900,9 +1893,9 @@ command_list_domains (void)
               err = gpg_error_from_syserror ();
               goto leave;
             }
-          if ((ec = gnupg_access (fname, W_OK)))
+          if (access (fname, W_OK))
             {
-              err = gpg_error (ec);
+              err = gpg_error_from_syserror ();
               if (gpg_err_code (err) == GPG_ERR_ENOENT)
                 {
                   if (gnupg_mkdir (fname, requireddirs[i].perm))
@@ -1930,9 +1923,9 @@ command_list_domains (void)
           err = gpg_error_from_syserror ();
           goto leave;
         }
-      if ((ec = gnupg_access (fname, F_OK)))
+      if (access (fname, F_OK))
         {
-          err = gpg_error (ec);
+          err = gpg_error_from_syserror ();
           if (gpg_err_code (err) == GPG_ERR_ENOENT)
             log_error ("domain %s: submission address not configured\n",
                        domain);
@@ -1953,17 +1946,7 @@ command_list_domains (void)
       if (!fp)
         {
           err = gpg_error_from_syserror ();
-          if (gpg_err_code (err) == GPG_ERR_ENOENT)
-            {
-              fp = es_fopen (fname, "w");
-              if (!fp)
-                log_error ("domain %s: can't create policy file: %s\n",
-                           domain, gpg_strerror (err));
-              else
-                es_fclose (fp);
-              fp = NULL;
-            }
-          else
+          if (gpg_err_code (err) != GPG_ERR_ENOENT)
             log_error ("domain %s: error in policy file: %s\n",
                        domain, gpg_strerror (err));
         }
@@ -1972,8 +1955,17 @@ command_list_domains (void)
           struct policy_flags_s policy;
           err = wks_parse_policy (&policy, fp, 0);
           es_fclose (fp);
+          if (!err)
+            {
+              struct policy_flags_s empty_policy;
+              memset (&empty_policy, 0, sizeof empty_policy);
+              if (!memcmp (&empty_policy, &policy, sizeof policy))
+                log_error ("domain %s: empty policy file\n", domain);
+            }
           wks_free_policy (&policy);
         }
+
+
     }
   err = 0;
 
@@ -2005,22 +1997,210 @@ command_cron (void)
 }
 
 
+/* Install a single key into the WKD by reading FNAME and extracting
+ * USERID.  */
+static gpg_error_t
+command_install_key (const char *fname, const char *userid)
+{
+  gpg_error_t err;
+  KEYDB_SEARCH_DESC desc;
+  estream_t fp = NULL;
+  char *addrspec = NULL;
+  char *fpr = NULL;
+  uidinfo_list_t uidlist = NULL;
+  uidinfo_list_t uid, thisuid;
+  time_t thistime;
+  char *huname = NULL;
+  int any;
+
+  addrspec = mailbox_from_userid (userid);
+  if (!addrspec)
+    {
+      log_error ("\"%s\" is not a proper mail address\n", userid);
+      err = gpg_error (GPG_ERR_INV_USER_ID);
+      goto leave;
+    }
+
+  if (!classify_user_id (fname, &desc, 1)
+      && (desc.mode == KEYDB_SEARCH_MODE_FPR
+          || desc.mode == KEYDB_SEARCH_MODE_FPR20))
+    {
+      /* FNAME looks like a fingerprint.  Get the key from the
+       * standard keyring.  */
+      err = wks_get_key (&fp, fname, addrspec, 0);
+      if (err)
+        {
+          log_error ("error getting key '%s' (uid='%s'): %s\n",
+                     fname, addrspec, gpg_strerror (err));
+          goto leave;
+        }
+    }
+  else /* Take it from the file */
+    {
+      fp = es_fopen (fname, "rb");
+      if (!fp)
+        {
+          err = gpg_error_from_syserror ();
+          log_error ("error reading '%s': %s\n", fname, gpg_strerror (err));
+          goto leave;
+        }
+    }
+
+  /* List the key so that we can figure out the newest UID with the
+   * requested addrspec.  */
+  err = wks_list_key (fp, &fpr, &uidlist);
+  if (err)
+    {
+      log_error ("error parsing key: %s\n", gpg_strerror (err));
+      err = gpg_error (GPG_ERR_NO_PUBKEY);
+      goto leave;
+    }
+  thistime = 0;
+  thisuid = NULL;
+  any = 0;
+  for (uid = uidlist; uid; uid = uid->next)
+    {
+      if (!uid->mbox)
+        continue; /* Should not happen anyway.  */
+      if (ascii_strcasecmp (uid->mbox, addrspec))
+        continue; /* Not the requested addrspec.  */
+      any = 1;
+      if (uid->created > thistime)
+        {
+          thistime = uid->created;
+          thisuid = uid;
+        }
+    }
+  if (!thisuid)
+    thisuid = uidlist;  /* This is the case for a missing timestamp.  */
+  if (!any)
+    {
+      log_error ("public key in '%s' has no mail address '%s'\n",
+                 fname, addrspec);
+      err = gpg_error (GPG_ERR_INV_USER_ID);
+      goto leave;
+    }
+
+  if (opt.verbose)
+    log_info ("using key with user id '%s'\n", thisuid->uid);
+
+  {
+    estream_t fp2;
+
+    es_rewind (fp);
+    err = wks_filter_uid (&fp2, fp, thisuid->uid, 1);
+    if (err)
+      {
+        log_error ("error filtering key: %s\n", gpg_strerror (err));
+        err = gpg_error (GPG_ERR_NO_PUBKEY);
+        goto leave;
+      }
+    es_fclose (fp);
+    fp = fp2;
+  }
+
+  /* Hash user ID and create filename.  */
+  err = compute_hu_fname (&huname, addrspec);
+  if (err)
+    goto leave;
+
+  /* Publish.  */
+  err = write_to_file (fp, huname);
+  if (err)
+    {
+      log_error ("copying key to '%s' failed: %s\n", huname,gpg_strerror (err));
+      goto leave;
+    }
+
+  /* Make sure it is world readable.  */
+  if (gnupg_chmod (huname, "-rwxr--r--"))
+    log_error ("can't set permissions of '%s': %s\n",
+               huname, gpg_strerror (gpg_err_code_from_syserror()));
+
+  if (!opt.quiet)
+    log_info ("key %s published for '%s'\n", fpr, addrspec);
+
+ leave:
+  xfree (huname);
+  free_uidinfo_list (uidlist);
+  xfree (fpr);
+  xfree (addrspec);
+  es_fclose (fp);
+  return err;
+}
+
+
+/* Return the filename and optionally the addrspec for USERID at
+ * R_FNAME and R_ADDRSPEC.  R_ADDRSPEC might also be set on error.  */
+static gpg_error_t
+fname_from_userid (const char *userid, char **r_fname, char **r_addrspec)
+{
+  gpg_error_t err;
+  char *addrspec = NULL;
+  const char *domain;
+  char *hash = NULL;
+  const char *s;
+  char shaxbuf[32]; /* Used for SHA-1 and SHA-256 */
+
+  *r_fname = NULL;
+  if (r_addrspec)
+    *r_addrspec = NULL;
+
+  addrspec = mailbox_from_userid (userid);
+  if (!addrspec)
+    {
+      if (opt.verbose)
+        log_info ("\"%s\" is not a proper mail address\n", userid);
+      err = gpg_error (GPG_ERR_INV_USER_ID);
+      goto leave;
+    }
+
+  domain = strchr (addrspec, '@');
+  log_assert (domain);
+  domain++;
+
+  /* Hash user ID and create filename.  */
+  s = strchr (addrspec, '@');
+  log_assert (s);
+  gcry_md_hash_buffer (GCRY_MD_SHA1, shaxbuf, addrspec, s - addrspec);
+  hash = zb32_encode (shaxbuf, 8*20);
+  if (!hash)
+    {
+      err = gpg_error_from_syserror ();
+      goto leave;
+    }
+
+  *r_fname = make_filename_try (opt.directory, domain, "hu", hash, NULL);
+  if (!*r_fname)
+    err = gpg_error_from_syserror ();
+  else
+    err = 0;
+
+ leave:
+  if (r_addrspec && addrspec)
+    *r_addrspec = addrspec;
+  else
+    xfree (addrspec);
+  xfree (hash);
+  return err;
+}
+
+
 /* Check whether the key with USER_ID is installed.  */
 static gpg_error_t
 command_check_key (const char *userid)
 {
-  gpg_err_code_t ec;
   gpg_error_t err;
   char *addrspec = NULL;
   char *fname = NULL;
 
-  err = wks_fname_from_userid (userid, 0, &fname, &addrspec);
+  err = fname_from_userid (userid, &fname, &addrspec);
   if (err)
     goto leave;
 
-  if ((ec = gnupg_access (fname, R_OK)))
+  if (access (fname, R_OK))
     {
-      err = gpg_error (ec);
+      err = gpg_error_from_syserror ();
       if (opt_with_file)
         es_printf ("%s n %s\n", addrspec, fname);
       if (gpg_err_code (err) == GPG_ERR_ENOENT)
@@ -2049,11 +2229,49 @@ command_check_key (const char *userid)
 }
 
 
+/* Remove the key with mail address in USERID.  */
+static gpg_error_t
+command_remove_key (const char *userid)
+{
+  gpg_error_t err;
+  char *addrspec = NULL;
+  char *fname = NULL;
+
+  err = fname_from_userid (userid, &fname, &addrspec);
+  if (err)
+    goto leave;
+
+  if (gnupg_remove (fname))
+    {
+      err = gpg_error_from_syserror ();
+      if (gpg_err_code (err) == GPG_ERR_ENOENT)
+        {
+          if (!opt.quiet)
+            log_info ("key for '%s' is not installed\n", addrspec);
+          log_inc_errorcount ();
+          err = 0;
+        }
+      else
+        log_error ("error removing '%s': %s\n", fname, gpg_strerror (err));
+      goto leave;
+    }
+
+  if (opt.verbose)
+    log_info ("key for '%s' removed\n", addrspec);
+  err = 0;
+
+ leave:
+  xfree (fname);
+  xfree (addrspec);
+  return err;
+}
+
+
 /* Revoke the key with mail address MAILADDR.  */
 static gpg_error_t
 command_revoke_key (const char *mailaddr)
 {
   /* Remove should be different from removing but we have not yet
    * defined a suitable way to do this.  */
-  return wks_cmd_remove_key (mailaddr);
+  return command_remove_key (mailaddr);
 }
